"""This file is for docstring-relevant functionalities."""


def format_docstring(**kwargs):
    """Decorator that formats the docstring with specified keywords.

    Notes
    -----
    If the docstring is to be formatted and it includes curly braces (for example, when
    typing literals), one must use double curly braces instead.
    """

    def wrapped(func_or_cls):
        func_or_cls.__doc__ = func_or_cls.__doc__.format(**kwargs)
        return func_or_cls

    return wrapped



import asyncio
import time
from functools import partial

import pytest

from ml3m._async import AsyncRunner
from ml3m.errors import InvalidParameterError

#######################################################################################
#                                                                                     #
#                                  PREPARATION WORK                                   #
#                                                                                     #
#######################################################################################


def process_func(item, **kwargs):
    """Pass all items."""
    return item * 10, [("Done", f"Succeeded on {item}")], None


async def process_afunc(item, addtlks=None, **kwargs):
    """Pass all items."""
    await asyncio.sleep(0.01)
    return process_func(item, **kwargs)


def failing_process_func(item, mode, **kwargs):
    """Fail items >= 5, supporting different modes of failing."""
    if item >= 5:
        if mode == "proper":
            return None, None, ("Done", f"Failed on {item}")
        elif mode == "exception":
            raise ValueError
        elif mode == "cannot_unpack":
            return item * 10, None
    return item * 10, [("Done", f"Succeeded on {item}")], None


async def failing_process_afunc(item, mode, addtlks=None, **kwargs):
    """Fail items >= 5, supporting different modes of failing."""
    await asyncio.sleep(0.01)
    return failing_process_func(item, mode, **kwargs)


def process_func_novarkey(item):
    """Accepts no variable-length keyword arguments."""


async def process_afunc_novarkey(item, addtlks=None):
    """Accepts no variable-length keyword arguments."""


async def process_afunc_noaddtlks(item, **kwargs):
    """Does not accept addtlks."""


async def process_afunc_addtlks_positional(item, addtlks, **kwargs):
    """Positional addtlks thus no default value."""


async def process_afunc_addtlks_default_not_none(item, addtlks=1, **kwargs):
    """Default value of addtlks not None."""


#######################################################################################
#                                                                                     #
#                                  TESTS START HERE                                   #
#                                                                                     #
#######################################################################################


class TestAsyncRunner:
    """Testing ml3m._async.AsyncRunner."""

    @pytest.mark.parametrize(
        "worker_kwargs",
        [
            [{}],
            [{"dummy": None}],
            [{}, {}],
            [{"dummy": None}, {"dummy": None}],
        ],
    )
    @pytest.mark.parametrize(
        "func,afunc,passed,failed",
        [(process_func, process_afunc, range(0, 100, 10), [])]
        + [
            (
                partial(failing_process_func, mode=mode),
                partial(failing_process_afunc, mode=mode),
                range(0, 50, 10),
                range(5, 10),
            )
            for mode in ["proper", "exception", "cannot_unpack"]
        ],
    )
    def test_async_runner_basics(self, func, afunc, passed, failed, worker_kwargs):
        """Test the basic functionalities of the asynchronous runner."""
        runner = AsyncRunner(process_func=func, process_afunc=afunc)

        results, failed_items = runner.run(
            items=list(range(10)), worker_kwargs=worker_kwargs
        )
        assert set(results) == set(passed)
        assert set(failed_items) == set(failed)

    def test_async_runner_speedup(self):
        """Test that asynchronous parallelization speeds up."""
        items = [0.01] * 100

        def process_func(item, **kwargs):
            time.sleep(item)
            return item, f"Slept {item}s", None

        async def process_afunc(item, addtlks=None, **kwargs):
            await asyncio.sleep(item)
            return item, f"Slept {item}s", None

        runner = AsyncRunner(process_func=process_func, process_afunc=process_afunc)

        # Running with only a single worker
        s = time.time()
        runner.run(items=items, worker_kwargs=[{}])
        single_worker_time = time.time() - s

        # Running with ten workers
        s = time.time()
        runner.run(items=items, worker_kwargs=[{} for _ in range(10)])
        multi_worker_time = time.time() - s

        # Loosen the speedup: 10 workers, at least 5x speedup
        assert multi_worker_time < single_worker_time / 5

    @pytest.mark.parametrize(
        "worker_kwargs,process_func,process_afunc",
        [
            ([{}], None, process_afunc),  # process_func is None, 1 worker
            ([{}], process_afunc, None),  # process_func is asynchronous
            ([{}, {}], process_func, None),  # process_afunc is None, >1 worker
            ([{}, {}], None, process_func),  # process_afunc is synchronous
            ([{}], process_func_novarkey, None),
            ([{}, {}], None, process_afunc_novarkey),
            ([{}, {}], None, process_afunc_noaddtlks),
            ([{}, {}], None, process_afunc_addtlks_positional),
            ([{}, {}], None, process_afunc_addtlks_default_not_none),
        ],
    )
    def test_async_runner_invalid_run(self, worker_kwargs, process_func, process_afunc):
        """Test invalid initialization and running arguments."""
        runner = AsyncRunner(
            process_func=process_func,
            process_afunc=process_afunc,
        )

        with pytest.raises(InvalidParameterError):
            runner.run(list(range(10)), worker_kwargs=worker_kwargs)



from __future__ import annotations

import json
import re
from typing import TYPE_CHECKING, Any, Callable

from .._metrics import bleu
from ..base.eval import BaseEvaluator, BaseOpenAIEvaluator
from ..errors import InvalidParameterError

if TYPE_CHECKING:
    from numbers import Real
    from pathlib import Path

    from .._typing import DataItemType, DatasetFormat, LoggingMode, QaSubject


class QaOpenAIEvaluator(BaseOpenAIEvaluator):
    """Evaluator for question-answering via OpenAI.

    This evaluator utilizes the ability of OpenAI models to tell the quality of a
    response from the following aspects:

    - **Accuracy**: Using the reference answer as the ground truth, does the response
      include factually incorrect information?
    - **Completeness**: Compared with the reference answer, is the response missing
      details?
    - **Clarity**: Is the response well-organized and clearly presented? If accuracy
      and completeness is poor, clarity should also be considered poor.

    Parameters
    ----------
    dataset : str or pathlib.Path
        The absolute path to the evaluation dataset.
    save_path : str or pathlib.Path
        The absolute path to the save location. This path may or may not exist, and if
        it exists, its file contents will be treated as a (partially) written result.
        Whether to overwrite the existing results or to build on them depend on
        ``overwrite`` when using the :meth:`QaOpenAIEvaluator.evaluate` method.
    openai_config : str or pathlib.Path
        The absolute path to the OpenAI configuration file.
    info_func : Callable
        The function that extracts the question, actual answer, and expected answer of
        a data item. The input parameter should be a :class:`pandas.Series`, a list, or
        a dictionary, depending on ``fmt`` and the specific type of each data item. The
        output should be a tuple of three strings, respectively the question, the actual
        answer to that question, and the expected answer of that question. See the notes
        for examples.
    fmt : {"jsonl", "json", "csv"}, default="jsonl"
        The format of ``dataset``.
    aspects : list of {"accuracy", "completeness", "clarity"}, optional
        The aspects to evaluate. If ``None``, evaluate all available aspects.
    setting: str, optional
        The personality setting for the OpenAI model, passed as the system message. If
        ``None``, then no system message is used.
    n_iter : int, default=3
        The number of iterations for each data item. The mean of the scores for each
        data item will be taken as the final score.
    timeout : float, default=60
        The timeout in seconds. This is not the OpenAI timeout, but the timeout for
        cancelling the worker tasks.
    model : str, default="gpt-3.5-turbo"
        The ID of the model to use, must be one of the available OpenAI models that
        support the ChatCompletion API. See also
        https://platform.openai.com/docs/models/model-endpoint-compatibility
    logging_mode : {"all", "failed", "none"}, default="all"
        The logging mode, whether to save the logs of all items, or only of failed
        items, or save no log.
    verbose : int, default=0
        The verbosity level of the processing. For negative levels, only a progress bar
        will be displayed. For level 0, the errored items will also be displayed. For
        positive levels, the all items will be displayed, and the verbosity level
        determines the number of lines to display for the message of each item.

    Notes
    -----
    Here are some examples of ``info_func``:

    Assume that ``dataset`` is in ``.jsonl`` format and each line is of the following
    form: ``{{"instruction": "xxx", "input": "xxx", "output": "xxx", "history": [],
    "response": "xxx"}}``. Then ``info_func`` can be defined as follows:

    .. code-block:: python

        def info_func(data_item: dict) -> tuple[str, str, str]:
            question = data_item["instruction"] + "\\n" + data_item["input"]
            actual = data_item["response"]
            expected = data_item["output"]
            return question, actual, expected

    Now assume that ``dataset`` is in ``.csv`` format with columns "question",
    "answer", and "response". Then ``info_func`` can be defined as follows:

    .. code-block:: python

        def info_func(data_item: pandas.Series) -> tuple[str, str, str]:
            question, answer, response = data_item[["question", "answer", "response"]]
            return question, response, answer
    """

    _pattern = re.compile(r"```[a-z]*\n(.+)\n```")

    def __init__(
        self,
        dataset: str | Path,
        save_path: str | Path,
        openai_config: str | Path,
        info_func: Callable[[DataItemType], tuple[str, str, str]],
        *,
        fmt: DatasetFormat = "jsonl",
        aspects: list[QaSubject] | None = None,
        setting: str | None = None,
        n_iter: int = 3,
        timeout: float = 60,
        model: str = "gpt-3.5-turbo",
        logging_mode: LoggingMode = "all",
        verbose: int = 0,
    ) -> None:
        self.info_func = info_func
        self.setting = setting

        # Determine the aspects to evaluate on
        avail_aspects: list[QaSubject] = ["accuracy", "completeness", "clarity"]
        self.aspects = avail_aspects if aspects is None else aspects

        # Validate the arguments
        if not callable(self.info_func):
            raise InvalidParameterError(
                "info_func", actual=self.info_func, reason="must be a callable"
            )
        if any(subject not in avail_aspects for subject in self.aspects) or len(
            self.aspects
        ) != len(set(self.aspects)):
            raise InvalidParameterError(
                "aspects",
                actual=self.aspects,
                reason=(
                    f"must be a list of non-duplicated aspects among {avail_aspects}"
                ),
            )

        # Set the subject explanations
        self._explanations: list[str] = []
        if "accuracy" in self.aspects:
            self._explanations.append(
                "accuracy: Using the reference answer as the ground truth, does my "
                "answer include factually incorrect information?"
            )
        if "completeness" in self.aspects:
            self._explanations.append(
                "completeness: Compared with the reference answer, is my answer "
                "missing details?"
            )
        if "clarity" in self.aspects:
            self._explanations.append(
                "clarity: Is my answer well-organized and clearly presented? If "
                "accuracy and completeness is bad, clarity should also be bad."
            )

        # Inherit from parent
        super().__init__(
            dataset=dataset,
            save_path=save_path,
            subjects=self.aspects,
            openai_config=openai_config,
            fmt=fmt,
            n_iter=n_iter,
            agg_method="mean",
            timeout=timeout,
            model=model,
            logging_mode=logging_mode,
            verbose=verbose,
        )

    def _prompt(self, data_item: DataItemType) -> tuple[str, str]:
        """:meta private:"""
        question, actual, expected = self.info_func(data_item)
        explanation_expr = "\n".join(
            [
                f"{i + 1}. {explanation}"
                for i, explanation in enumerate(self._explanations)
            ]
        )
        return (
            "" if self.setting is None else self.setting,
            f"### Question\n```\n{question}\n```\n\n### My answer\n```\n{actual}\n```"
            f"\n\n### Reference answer\n```\n{expected}\n```\n\n### Scoring\n\nI want "
            "you to score my answer based on the reference answer in the following "
            f"aspects:\n{explanation_expr}\n\nEach score should be from 1 to 5. Be "
            "very strict. In your response, you should only include a JSON object, "
            "with keys being the aspects and values being the scores. Do not include "
            "any additional information or explanation.",
        )

    def _extract_scores(
        self, reply: str, data_item: DataItemType
    ) -> Real | dict[Any, Real]:
        """:meta private:"""
        scores: dict[QaSubject, Real]
        try:
            scores = {
                subject.lower(): score for subject, score in json.loads(reply).items()
            }

        # Try to search for a code block in the reply; if errored, leave as is
        except:
            match = re.search(self._pattern, reply)
            assert match is not None, "backup pattern matching failed."
            scores = {
                subject.lower(): score
                for subject, score in json.loads(match.group(1)).items()
            }
        return scores


class QaMetricEvaluator(BaseEvaluator):
    """Evaluator for question-answering via common metrics.

    This evaluator supports using the following metric to compare the actual response
    with the reference answer:

    - `BLEU-k <https://en.wikipedia.org/wiki/BLEU>`_ (BiLingual Evaluation Understudy)

    Parameters
    ----------
    dataset : str or pathlib.Path
        The absolute path to the evaluation dataset.
    save_path : str or pathlib.Path
        The absolute path to the save location. This path may or may not exist, and if
        it exists, its file contents will be treated as a (partially) written result.
        Whether to overwrite the existing results or to build on them depend on
        ``overwrite`` when using the :meth:`QaMetricEvaluator.evaluate` method.
    info_func : Callable
        The function that extracts the question, actual answer, and expected answer of
        a data item. The input parameter should be a :class:`pandas.Series`, a list, or
        a dictionary, depending on ``fmt`` and the specific type of each data item. The
        output should be a tuple of three strings, respectively the question, the actual
        answer to that question, and the expected answer of that question. See the notes
        for examples.
    fmt : {"jsonl", "json", "csv"}, default="jsonl"
        The format of ``dataset``.
    bleu_k : list of int or None
        The list of k-values used for BLEU-k. Must be positive. If ``None``, use k-
        values 1 to 4.
    logging_mode : {"all", "failed", "none"}, default="all"
        The logging mode, whether to save the logs of all items, or only of failed
        items, or save no log.
    verbose : int, default=0
        The verbosity level of the processing. For negative levels, only a progress bar
        will be displayed. For level 0, the errored items will also be displayed. For
        positive levels, the all items will be displayed, and the verbosity level
        determines the number of lines to display for the message of each item.

    Notes
    -----
    Here are some examples of ``info_func``:

    Assume that ``dataset`` is in ``.jsonl`` format and each line is of the following
    form: ``{{"instruction": "xxx", "input": "xxx", "output": "xxx", "history": [],
    "response": "xxx"}}``. Then ``info_func`` can be defined as follows:

    .. code-block:: python

        def info_func(data_item: dict) -> tuple[str, str, str]:
            question = data_item["instruction"] + "\\n" + data_item["input"]
            actual = data_item["response"]
            expected = data_item["output"]
            return question, actual, expected

    Now assume that ``dataset`` is in ``.csv`` format with columns "question",
    "answer", and "response". Then ``info_func`` can be defined as follows:

    .. code-block:: python

        def info_func(data_item: pandas.Series) -> tuple[str, str, str]:
            question, answer, response = data_item[["question", "answer", "response"]]
            return question, response, answer
    """

    def __init__(
        self,
        dataset: str | Path,
        save_path: str | Path,
        info_func: Callable[[DataItemType], tuple[str, str, str]],
        *,
        fmt: DatasetFormat = "jsonl",
        bleu_k: list[int] | None = None,
        logging_mode: LoggingMode = "all",
        verbose: int = 0,
    ) -> None:
        self.info_func = info_func
        self.bleu_k = [1, 2, 3, 4] if bleu_k is None else bleu_k

        # Validate the arguments
        if not isinstance(self.bleu_k, list) or any(
            not isinstance(k, int) or k <= 0 for k in self.bleu_k
        ):
            raise InvalidParameterError(
                "bleu_k",
                actual=self.bleu_k,
                reason="must be a list of positive integers",
            )

        # Inherit from parent
        super().__init__(
            dataset=dataset,
            save_path=save_path,
            subjects=[f"BLEU-{k}" for k in self.bleu_k],
            fmt=fmt,
            workers=1,
            n_iter=1,
            agg_method=None,
            logging_mode=logging_mode,
            verbose=verbose,
        )

    def _get_score(self, data_item: DataItemType, **kwargs) -> Real | dict[Any, Real]:
        """:meta private:"""
        _, actual, expected = self.info_func(data_item)
        scores = {}

        # Compute the desired BLEU scores
        for k, bleu_score in zip(self.bleu_k, bleu(actual, expected, self.bleu_k)):
            scores[f"BLEU-{k}"] = bleu_score

        # mypy not working with numbers.Real
        return scores  # type: ignore[return-value]

    async def _aget_score(
        self, data_item: DataItemType, **kwargs
    ) -> Real | dict[Any, Real]:
        """:meta private:"""
        return self._get_score(data_item, **kwargs)



"""This file is for asynchronous parallelization functionalities."""


from __future__ import annotations

import asyncio
from inspect import _ParameterKind, iscoroutinefunction, signature
from typing import Any, Callable, Iterable, NoReturn

from tqdm import tqdm

from ._display import COLOR, EMOJI, colored, wrap_with_prefix
from .errors import InvalidParameterError


class AsyncRunner:
    """An asynchronous runner.

    Parameters
    ----------
    process_func : Callable, optional
        The processing function that takes a data item then returns three things: the
        result, a normal message, and an error message. If the processing succeeds, the
        error message should be ``None``. If the processing errors out, the result and
        the normal message should be ``None``. The normal message should be a list of
        prefix-content pairs. The error message should be a prefix-content pair.

        .. note::
            This function must be provided if the runner is intended to be run with a
            single worker.

        .. note::
            This function needs to accept variable-length keyword arguments. Also, this
            function is not supposed to raise any exception (though they will still be
            handled to avoid potential errors). Instead, all exceptions should be
            caught and reflected through the result, the normal message, and the error
            message as specified above.
    process_afunc : Callable, optional
        The asynchronous version of ``process_func``.

        .. note::
            This function must be provided if the runner is intended to be run with
            multiple workers.
        .. note::
            This function must be asynchronous. Also, in addition to accepting
            variable-length keyword arguments, it also needs to accept a keyword
            argument ``addtlks`` of type ``list[asyncio.Lock] | None`` with default
            value ``None``. It should also handle all possible exceptions, for the same
            reason as described in ``process_func``.
    verbose : int, default=0
        The verbosity level of the processing. For negative levels, only a progress bar
        will be displayed. For level 0, the errored items will also be displayed. For
        positive levels, the all items will be displayed, and the verbosity level
        determines the number of lines to display for the message of each item.
    """

    def __init__(
        self,
        process_func: Callable | None = None,
        process_afunc: Callable | None = None,
        verbose: int = 0,
    ):
        self.process_func = process_func
        self.process_afunc = process_afunc
        self.verbose = verbose

    def _validate_process_function(self, func: Callable, async_version: bool) -> None:
        """Validate the processing function.

        Parameters
        ----------
        func : Callable
            The function to evaluate.
        async_version : bool
            Whether the function to evaluate is the asynchronous processing function or
            the synchronous one.
        """
        if async_version ^ iscoroutinefunction(func):
            raise InvalidParameterError(
                "process_(a)func",
                actual=func,
                reason="must be (a)synchronous",
            )

        # Check the function signature
        sig = signature(func)
        if async_version and "addtlks" not in sig.parameters:
            raise InvalidParameterError(
                "process_afunc",
                actual=func,
                reason="must accept a keyword argument 'addtlks'",
            )
        elif async_version and sig.parameters["addtlks"].default is not None:
            raise InvalidParameterError(
                "process_afunc",
                actual=func,
                reason="keyword argument 'addtlks' must be defaulted to 'None'",
            )
        if not any(
            param.kind == _ParameterKind.VAR_KEYWORD
            for param in sig.parameters.values()
        ):
            raise InvalidParameterError(
                "process_(a)func",
                actual=func,
                reason="must accept variable-length **kwargs",
            )

    def _sequential(self) -> None:
        """Sequential processing."""
        all_items = list(self.items)
        self.progbar = tqdm(total=len(all_items))

        # Process each item sequentially
        for item in all_items:
            result: Any | None
            norm_msg: list[tuple[Any, Any]] | None
            err_msg: tuple[Any, Any] | None

            # Avoid explicitly raising the exception that interrupts the runner
            try:
                # Validated self.process_func is not None before calling
                result, norm_msg, err_msg = self.process_func(  # type: ignore[misc]
                    item, **self.worker_kwargs[0]
                )
                assert (
                    norm_msg is None
                    or isinstance(norm_msg, list)
                    and all(
                        isinstance(item, tuple) and len(item) == 2 for item in norm_msg
                    )
                ), "invalid norm_msg type"
                assert (
                    err_msg is None or isinstance(err_msg, tuple) and len(err_msg) == 2
                ), "invalid err_msg type"
            except Exception as e:
                result, norm_msg = None, None
                err_msg = ("UNEXPECTED", f"{type(e).__name__}: {e}")

            # Print to console according to verbosity level
            if result is None and err_msg is not None and self.verbose >= 0:
                tqdm.write(
                    wrap_with_prefix(
                        err_msg[0],
                        err_msg[1],
                        max_lines=self.verbose,
                        prefix_color=COLOR.RED,
                        content_color=COLOR.RED,
                    )
                )
            elif result is not None and norm_msg is not None and self.verbose >= 1:
                to_console = [
                    wrap_with_prefix(
                        norm_prefix,
                        norm_content,
                        max_lines=self.verbose,
                        prefix_color=COLOR.GREEN,
                    )
                    for norm_prefix, norm_content in norm_msg
                ]
                tqdm.write("\n".join(to_console) + "\n")
            self.progbar.update(1)

            # Collect the result
            self.shared_resources.append(
                (True, result) if result is not None else (False, item)
            )
        self.progbar.close()

    async def _mainloop(self) -> None:
        """Main event loop for asynchronous parallelization."""
        self.queue: asyncio.Queue[Any] = asyncio.Queue()
        n_items = 0
        for item in self.items:
            self.queue.put_nowait(item)
            n_items += 1
        if n_items == 0:
            return

        # Create necessary asynchronous locks and additional ones on demand
        self.mainlk = asyncio.Lock()
        self.proglk = asyncio.Lock()
        self.addtlks = [asyncio.Lock() for _ in range(self.n_locks)]

        # Create worker tasks to process the queue asynchronously
        print(f"Initializing {len(self.worker_kwargs)} workers for {n_items} items...")
        tasks: list[asyncio.Task] = []
        self.progbar = tqdm(total=n_items)
        for worker_id, kwargs in enumerate(self.worker_kwargs):
            tasks.append(asyncio.create_task(self._worker(worker_id, **kwargs)))

        # Wait until the queue is fully processed and collect the results
        await self.queue.join()
        for task in tasks:
            task.cancel()
        await asyncio.gather(*tasks, return_exceptions=True)
        self.progbar.close()

    async def _worker(self, worker_id: int, **kwargs) -> NoReturn:
        """The worker for processing the asynchronous queue.

        Parameters
        ----------
        worker_id : int
            The id of the worker.
        kwargs
            The additional keyword arguments.
        """
        while True:
            item = await self.queue.get()
            result: Any | None
            norm_msg: list[tuple[Any, Any]] | None
            err_msg: tuple[Any, Any] | None

            # Avoid "task exception never being retrieved"
            try:
                # Validated self.process_afunc is not None before calling
                (
                    result,
                    norm_msg,
                    err_msg,
                ) = await self.process_afunc(  # type: ignore[misc]
                    item, addtlks=self.addtlks, **kwargs
                )
                assert (
                    norm_msg is None
                    or isinstance(norm_msg, list)
                    and all(
                        isinstance(item, tuple) and len(item) == 2 for item in norm_msg
                    )
                ), "invalid norm_msg type"
                assert (
                    err_msg is None or isinstance(err_msg, tuple) and len(err_msg) == 2
                ), "invalid err_msg type"
            except Exception as e:
                result, norm_msg = None, None
                err_msg = ("UNEXPECTED", f"{type(e).__name__}: {e}")

            # Print to console according to verbosity level
            async with self.proglk:
                if result is None and err_msg is not None and self.verbose >= 0:
                    to_console = wrap_with_prefix(
                        err_msg[0],
                        err_msg[1],
                        max_lines=self.verbose,
                        prefix_color=COLOR.RED,
                        content_color=COLOR.RED,
                    )
                    if self.verbose >= 1:
                        to_console += "\n"
                    tqdm.write(to_console)
                elif result is not None and norm_msg is not None and self.verbose >= 1:
                    to_console_list = [
                        wrap_with_prefix(
                            norm_prefix,
                            norm_content,
                            max_lines=self.verbose,
                            prefix_color=COLOR.GREEN,
                        )
                        for norm_prefix, norm_content in norm_msg[:-1]
                    ] + [
                        wrap_with_prefix(
                            norm_msg[-1][0],
                            norm_msg[-1][1],
                            max_lines=self.verbose,
                            prefix_color=COLOR.GREEN,
                            content_color=COLOR.GREEN,
                        )
                    ]
                    tqdm.write("\n".join(to_console_list) + "\n")
                self.progbar.update(1)

            # Collect the result and mark the task as done
            async with self.mainlk:
                self.shared_resources.append(
                    (True, result) if result is not None else (False, item)
                )
            self.queue.task_done()

    def run(
        self, items: Iterable, worker_kwargs: list[dict[str, Any]], n_locks: int = 0
    ) -> tuple[list, list]:
        """Asynchronously (or sequentially) process the items.

        Parameters
        ----------
        items : Iterable
            The items to process with either ``process_func`` or ``process_afunc``,
            depending on the number of workers.
        worker_kwargs : list of dict
            The additional keyword arguments to pass into ``process_func`` or
            ``process_afunc``. The length of ``worker_kwargs`` determines the number of
            workers to create, thus also determining whether to run in asynchronous or
            sequential mode.
        n_locks : int, default=0
            The additional locks to request. The requested locks will be passed to
            ``process_afunc`` via the keyword argument ``addtlks``. Not useful when
            running in sequential mode.

        Returns
        -------
        results : list
            The successfully processed results.
        failed_items : list
            The failed items.
        """
        self.items = items
        self.worker_kwargs = worker_kwargs
        self.n_locks = n_locks
        self.shared_resources: list[tuple[bool, Any]] = []

        # Update the shared resources, either in synchronous or asynchronous mode
        if len(self.worker_kwargs) > 1:
            if self.process_afunc is None:
                raise InvalidParameterError(
                    "process_afunc",
                    actual=self.process_afunc,
                    reason="must be defined when using multiple workers",
                )
            self._validate_process_function(self.process_afunc, async_version=True)
            asyncio.run(self._mainloop())
        else:
            if self.process_func is None:
                raise InvalidParameterError(
                    "process_func",
                    actual=self.process_func,
                    reason="must be defined when using a single worker",
                )
            self._validate_process_function(self.process_func, async_version=False)
            self._sequential()

        # Collect finished results and failed items
        print("Collecting results...", end=" ", flush=True)
        results: list = []
        failed_items: list = []
        for passed, obj in self.shared_resources:
            if passed:
                results.append(obj)
            else:
                failed_items.append(obj)

        # Print a short summary of the execution
        print(
            f"Done/Failed - {colored(len(results), COLOR.GREEN)}/"
            f"{colored(len(failed_items), COLOR.RED)}"
        )
        if not failed_items:
            print(f"{EMOJI.STAR} All items have been successfully processed!")
        return results, failed_items



from __future__ import annotations

import random
import re
from typing import TYPE_CHECKING, Any, Callable, Literal

from ..base.eval import BaseOpenAIEvaluator
from ..errors import InvalidParameterError

if TYPE_CHECKING:
    from numbers import Real
    from pathlib import Path

    from .._typing import DataItemType, DatasetFormat, LoggingMode


class McqOpenAIEvaluator(BaseOpenAIEvaluator):
    """Evaluator for multiple-choice questions via OpenAI.

    This evaluator utilizes the ability of OpenAI models to tell if a response selects
    the correct options, based on the reference answer. The score for each data item
    would be either 0 or 100, and there will be no partial credits.

    Parameters
    ----------
    dataset : str or pathlib.Path
        The absolute path to the evaluation dataset.
    save_path : str or pathlib.Path
        The absolute path to the save location. This path may or may not exist, and if
        it exists, its file contents will be treated as a (partially) written result.
        Whether to overwrite the existing results or to build on them depend on
        ``overwrite`` when using the :meth:`McqOpenAIEvaluator.evaluate` method.
    openai_config : str or pathlib.Path
        The absolute path to the OpenAI configuration file.
    info_func : Callable
        The function that extracts the question, actual answer, and expected answer of
        a data item. The input parameter should be a :class:`pandas.Series`, a list, or
        a dictionary, depending on ``fmt`` and the specific type of each data item. The
        output should be a tuple of three strings, respectively the question, the actual
        answer to that question, and the expected answer of that question. See the notes
        for examples.
    fmt : {"jsonl", "json", "csv"}, default="jsonl"
        The format of ``dataset``.
    score_name : str, default="score"
        The key/column name to use for the obtained score. This should *not* be a key
        or column name that already exists in the save location. Be extremely careful
        since there will be *no* warning or exception raised on this.
    label_type : {"upper", "lower", "digit"}, default="upper"
        The type of the option labels. "upper" stands for A, B, C, D, ... "lower"
        stands for a, b, c, d, ... "digit" stands for 1, 2, 3, 4, ...
    label_cnt : int, default=4
        The number of options. For instance, ``label_type="upper"`` with
        ``label_cnt=4`` means that the option labels are A, B, C, and D.
    setting: str, optional
        The personality setting for the OpenAI model, passed as the system message. If
        ``None``, then no system message is used.
    n_iter : int, default=1
        The number of iterations for each data item. The mode of the scores for each
        data item will be taken as the final score.
    timeout : float, default=60
        The timeout in seconds. This is not the OpenAI timeout, but the timeout for
        cancelling the worker tasks.
    model : str, default="gpt-3.5-turbo"
        The ID of the model to use, must be one of the available OpenAI models that
        support the ChatCompletion API. See also
        https://platform.openai.com/docs/models/model-endpoint-compatibility
    logging_mode : {"all", "failed", "none"}, default="all"
        The logging mode, whether to save the logs of all items, or only of failed
        items, or save no log.
    verbose : int, default=0
        The verbosity level of the processing. For negative levels, only a progress bar
        will be displayed. For level 0, the errored items will also be displayed. For
        positive levels, the all items will be displayed, and the verbosity level
        determines the number of lines to display for the message of each item.

    Notes
    -----
    Here are some examples of ``info_func``:

    Assume that ``dataset`` is in ``.jsonl`` format and each line is of the following
    form: ``{{"instruction": "xxx", "input": "xxx", "output": "xxx", "history": [],
    "response": "xxx"}}``. Then ``info_func`` can be defined as follows:

    .. code-block:: python

        def info_func(data_item: dict) -> tuple[str, str, str]:
            question = data_item["instruction"] + "\\n" + data_item["input"]
            actual = data_item["response"]
            expected = data_item["output"]
            return question, actual, expected

    Now assume that ``dataset`` is in ``.csv`` format with columns "question", "A",
    "B", "C", "D", "answer", and "response". Then ``info_func`` can be defined as
    follows:

    .. code-block:: python

        def info_func(data_item: pandas.Series) -> tuple[str, str, str]:
            question, A, B, C, D, answer, response = data_item[
                ["question", "A", "B", "C", "D", "answer", "response"]
            ]
            formatted_question = (
                f"{{question}}\\nA. {{A}}\\nB. {{B}}\\nC. {{C}}\\nD. {{D}}"
            )
            return formatted_question, response, answer
    """

    def __init__(
        self,
        dataset: str | Path,
        save_path: str | Path,
        openai_config: str | Path,
        info_func: Callable[[DataItemType], tuple[str, str, str]],
        *,
        fmt: DatasetFormat = "jsonl",
        score_name: str = "score",
        label_type: Literal["upper", "lower", "digit"] = "upper",
        label_cnt: int = 4,
        setting: str | None = None,
        n_iter: int = 1,
        timeout: float = 60,
        model: str = "gpt-3.5-turbo",
        logging_mode: LoggingMode = "all",
        verbose: int = 0,
    ) -> None:
        self.info_func = info_func
        self.score_name = score_name
        self.label_type = label_type
        self.label_cnt = label_cnt
        self.setting = setting

        # Validate the arguments
        if not callable(self.info_func):
            raise InvalidParameterError(
                "info_func",
                actual=self.info_func,
                reason="must be a callable",
            )
        if label_type not in ["upper", "lower", "digit"]:
            raise InvalidParameterError(
                "label_type",
                actual=self.label_type,
                reason="must be one of 'upper', 'lower', and 'digit'.",
            )
        if not isinstance(label_cnt, int) or label_cnt <= 1:
            raise InvalidParameterError(
                "label_cnt",
                actual=self.label_cnt,
                reason="must be an integer > 1",
            )

        # Determine the actual labels
        self.labels: list[str]
        if label_type == "upper":
            if self.label_cnt > 13:
                raise ValueError("'label_type=upper' supports at most 13 labels.")
            self.labels = [chr(65 + i) for i in range(self.label_cnt)]
        elif label_type == "lower":
            if self.label_cnt > 26:
                raise ValueError("'label_type=lower' supports at most 26 labels.")
            self.labels = [chr(97 + i) for i in range(self.label_cnt)]
        else:  # label_type == "digit"
            self.labels = [str(i + 1) for i in range(self.label_cnt)]

        # Prepare for querying and reply processing
        label_grp = "|".join(self.labels)
        label_spls = [
            "".join(sorted(random.sample(self.labels, i + 1)))
            for i in range(min(3, self.label_cnt))
        ]
        self._few_shot = ", ".join(f"'{spl}'" for spl in label_spls)
        self._pat = re.compile(
            rf"selected option(?:\(s\)|s)? (is|are|is/are):? ([{label_grp}]+)"
        )

        # Inherit from parent
        super().__init__(
            dataset=dataset,
            save_path=save_path,
            subjects=[self.score_name],
            openai_config=openai_config,
            fmt=fmt,
            n_iter=n_iter,
            agg_method="mode",
            timeout=timeout,
            model=model,
            logging_mode=logging_mode,
            verbose=verbose,
        )

    def _prompt(self, data_item: DataItemType) -> tuple[str, str]:
        """:meta private:"""
        question, actual, _ = self.info_func(data_item)
        return (
            "" if self.setting is None else self.setting,
            "In this task, I will provide you a multiple-choice question and a "
            "student's answer to it. I want you to tell which options the student has "
            "selected as the solution to the multiple-choice question. The criteria "
            "are as follows:\n- If the student's answer explicitly mentions which "
            "are correct and which are wrong, he selects the correct ones.\n- If the "
            "student's answer only lists several (or all) options without saying "
            "whether they are correct or wrong, he selects all options he has listed."
            "\n- If none of the above criteria applies to the student's answer, judge "
            "from the content of his answer.\n\nThe multiple-choice question is:\n```"
            f"\n{question}\n```\n\nThe student's answer is:\n```\n{actual}\n```\n\n"
            "Remember that when you output the student's selected options, only "
            f"include the labels in the format e.g. {self._few_shot}. If the student "
            "selects none of the options, output only 'N'. Do not include any "
            "explanation or addidtional information! The student's selected option(s) "
            "is/are:",
        )

    def _extract_scores(
        self, reply: str, data_item: DataItemType
    ) -> Real | dict[Any, Real]:
        """:meta private:"""
        stripped_reply = reply.strip()

        # Try to match the reply pattern in advance if possible
        mat = re.search(self._pat, stripped_reply)
        if mat is not None:
            stripped_reply = mat.group(2)

        # Construct the set of chosen options
        chosen_options: set[str] = set()
        for char in stripped_reply:
            if not char.isspace() and char not in ",.，、" and char not in self.labels:
                raise ValueError(
                    f"Got invalid character '{char}' in '{stripped_reply}'."
                )
            chosen_options.add(char)

        # Compare with the reference answer
        _, _, expected = self.info_func(data_item)
        expected_options: set[str] = set()
        for char in expected:
            if not char.isspace() and char not in ",.，、" and char not in self.labels:
                raise ValueError(
                    f"[FATAL] Got invalid reference answer '{expected}' in the dataset."
                )
            expected_options.add(char)
        # mypy not working with numbers.Real
        return (expected_options == chosen_options) * 100  # type: ignore[return-value]



from __future__ import annotations

import shutil
import sys
from enum import Enum
from typing import TYPE_CHECKING, Any

import pandas as pd

if TYPE_CHECKING:
    from ._typing import DataItemType

#######################################################################################
#                                                                                     #
#                                        COLOR                                        #
#                                                                                     #
#######################################################################################


class COLOR(Enum):
    BLACK = "\033[30m"
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"
    DEFAULT = "\033[39m"


_color_support = True


if sys.platform == "win32":
    try:
        # https://stackoverflow.com/questions/36760127
        from ctypes import windll

        kernel32 = windll.kernel32
        kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), 7)
    except Exception:  # pragma: no cover
        _color_support = False


def colored(content: Any, color: COLOR | None) -> str:
    """Return the content that will be printed colored.

    Parameters
    ----------
    content : str
        The content to color.
    color : COLOR
        The color type.

    Returns
    -------
    colored_content : str
        The colored content. Default color if coloring is not supported.
    """
    if _color_support and color is not None:
        return f"{color.value}{content}{COLOR.DEFAULT.value}"
    return f"{content}"


#######################################################################################
#                                                                                     #
#                                        EMOJI                                        #
#                                                                                     #
#######################################################################################


class EMOJI:
    STAR = "\U0001F31F"
    DIAMOND = "\U0001F539"


#######################################################################################
#                                                                                     #
#                                    TEXT WRAPPING                                    #
#                                                                                     #
#######################################################################################


def wrap_text(text: str, width: int, max_lines: int | None = None) -> list[str]:
    """Text wrapper similar to textwrap.wrap.

    This considers any characters with order >= 0x1100 as two-characters-wide, because
    commonly this is used for printing to console and consoles commonly use monospace
    fonts where 2:1 is a close approximation. (Note that 0x1100 is the first wide
    character, though there are normal-width characters after that.)

    There might be still a slight overflow in the actual display due to the
    approximation. One would need to relax ``width`` if a strict guarantee is desired.

    Parameters
    ----------
    text : str
        The text for which its representation will be wrapped.
    width : int
        The maximum width of the wrapped text. Must be at least 7.
    max_lines : int, optional
        The maximum number of lines to display. If ``None``, there is no limit. If the
        text is too long to display within ``max_lines``, it will be marked with a
        placeholder "[...]".

    Returns
    -------
    wrapped : list of str
        The list of the wrapped lines of the text.
    """
    contents = list(repr(text)[-2:0:-1])
    if max_lines is not None:
        max_lines = max(1, max_lines)
    remaining_lines = max_lines
    wrapped: list[str] = []

    # Cannot accept width < 7 due to the placeholder
    if width < 7:
        raise ValueError("Require 'width' at least 7.")

    while contents:
        # Break early if no remaining lines are left and plug in placeholder
        if remaining_lines == 0:
            lastln, n_replace, cum_length = wrapped[-1], 0, 0
            while cum_length < 6:
                n_replace += 1
                cum_length += 2 if ord(lastln[-n_replace]) >= 0x1100 else 1
            wrapped[-1] = f"{lastln[:-n_replace]} [...]"
            break

        # Create a new line and loop until the next character cannot fit in
        line, line_length = [], 0
        while contents:
            char = contents[-1]
            char_length = 2 if ord(char) >= 0x1100 else 1
            if line_length + char_length > width:
                break
            line.append(char)
            line_length += char_length
            contents.pop()
        wrapped.append("".join(line))

        # Reduce the number of remaining lines if there is a restriction
        if remaining_lines is not None:
            remaining_lines -= 1

    return wrapped


def wrap_with_prefix(
    prefix: Any,
    content: Any,
    max_lines: int | None = None,
    prefix_color: COLOR | None = None,
    content_color: COLOR | None = None,
) -> str:
    """Wraps the content with a prefix in the first line.

    Parameters
    ----------
    prefix : Any
        The prefix, taking 1/4 of the terminal width from the left.
    content : Any
        The content, taking 3/4 of the terminal width from the right.
    max_lines : int, optional
        The maximum number of lines to display. If ``None``, there is no limit. The
        prefix and content, if too long to display within ``max_lines``, will be marked
        with a placeholder "[...]".
    prefix_color : COLOR, optional
        The color of the prefix. If ``None``, do not color the prefix.
    content_color : COLOR, optional
        The color of the content. If ``None``, do not color the content.

    Returns
    -------
    wrapped : str
        The wrapped (and colored) prefix and content.
    """
    prefix_width = shutil.get_terminal_size().columns // 4 - 2  # Avoid overflow
    content_width = prefix_width * 3

    # Make sure the maximum widths are enough for the placeholder
    prefix_width = max(7, prefix_width)
    content_width = max(7, content_width)

    # Wrap the prefix and the content respectively
    prefix_lns = wrap_text(f"{prefix}", width=prefix_width, max_lines=max_lines)
    content_lns = wrap_text(f"{content}", width=content_width, max_lines=max_lines)
    prefix_nlns, content_nlns = len(prefix_lns), len(content_lns)

    # Combine the prefix and the content lines
    lns = []
    for i in range(max(prefix_nlns, content_nlns)):
        prefix_ln = (
            colored("│", prefix_color) + " " * (prefix_width - 1)
            if i >= prefix_nlns
            else colored(f"{prefix_lns[i]:<{prefix_width}}", prefix_color)
        )
        content_ln = "" if i >= content_nlns else colored(content_lns[i], content_color)
        lns.append(f"{prefix_ln}  {content_ln}")
    return "\n".join(lns)


#######################################################################################
#                                                                                     #
#                                  OTHER FORMATTING                                   #
#                                                                                     #
#######################################################################################


def format_data_item(data_item: DataItemType) -> str:
    """Format a data item into a nice printout.

    Parameters
    ----------
    data_item : DataItemType
        The data item to format.

    Returns
    -------
    formatted_data_item : str
        The formatted data item.
    """
    if isinstance(data_item, list):
        return f" {EMOJI.DIAMOND} ".join(data_item)
    if isinstance(data_item, (dict, pd.Series)):
        return f" {EMOJI.DIAMOND} ".join([f"[{k}] {v}" for k, v in data_item.items()])
    raise TypeError(f"Invalid data item of type {type(data_item)}.")



from __future__ import annotations

import json
import os
import traceback
from datetime import datetime
from inspect import iscoroutinefunction
from typing import TYPE_CHECKING, Any, Callable, Generator

import pandas as pd

from .._async import AsyncRunner
from .._display import COLOR, colored
from .._logging import manage_timed_logs
from .._paths import ensure_path, validate_path
from ..errors import InvalidParameterError

if TYPE_CHECKING:
    import asyncio
    from pathlib import Path

    from .._typing import DataItemType, DatasetFormat, LoggingMode


class ResponseGenerator:
    """Generate responses and combine with the original dataset.

    Parameters
    ----------
    orig_dataset : str or pathlib.Path
        The absolute path to the original dataset.
    dataset : str or pathlib.Path
        The absolute path to the result dataset. All information in the original
        dataset will be preserved while the responses will be appended.
    info_func : Callable
        The function that takes a data item and forms the query. The data item can be a
        :class:`pandas.Series`, a list, or a dictionary, depending on ``format``.
        Whatever it returns will be passed as the input to ``query_func`` and printed
        to console for high verbosity levels.
    query_func : Callable
        The function that takes the query returned by ``info_func`` and outputs the
        model response represented as a single string. This function should be
        synchronous if ``n_workers=1`` and asynchronous otherwise.
    response_name : str
        The key or column name to use for the response. This should *not* be a key or
        column name that already exists in the dataset. Be extremely careful since
        there will be *no* warning or exception raised on this.
    fmt : {"jsonl", "json", "csv"}, default="jsonl"
        The format of ``dataset``.
    n_workers : int, default=1
        The number of workers. If only one worker, the dataset will be processed
        sequentially. Otherwise it will be asynchronously parallelized with the
        specified number of workers.
    logging_mode : {"all", "failed", "none"}, default="all"
        The logging mode, whether to save the logs of all items, or only of failed
        items, or save no log.
    verbose : int, default=0
        The verbosity level of the processing. For negative levels, only a progress bar
        will be displayed. For level 0, the errored items will also be displayed. For
        positive levels, the all items will be displayed, and the verbosity level
        determines the number of lines to display for the message of each item.
    """

    def __init__(
        self,
        orig_dataset: str | Path,
        dataset: str | Path,
        info_func: Callable[[DataItemType], Any],
        query_func: Callable[[Any], str],
        response_name: str,
        *,
        fmt: DatasetFormat = "jsonl",
        n_workers: int = 1,
        logging_mode: LoggingMode = "all",
        verbose: int = 0,
    ) -> None:
        self.orig_dataset = orig_dataset
        self.dataset = dataset
        self.info_func = info_func
        self.query_func = query_func
        self.fmt = fmt
        self.response_name = response_name
        self.n_workers = n_workers
        self.logging_mode = logging_mode
        self.verbose = verbose

        # Validate the arguments
        validate_path(self.orig_dataset)
        if not isinstance(self.n_workers, int) or self.n_workers < 1:
            raise InvalidParameterError(
                "n_workers",
                actual=self.n_workers,
                reason="must be an integer >= 1",
            )
        if not callable(self.info_func):
            raise InvalidParameterError(
                "info_func",
                actual=self.info_func,
                reason="must be a callable",
            )
        if not callable(self.query_func):
            raise InvalidParameterError(
                "query_func",
                actual=self.query_func,
                reason="must be a callable",
            )
        if self.n_workers == 1 and iscoroutinefunction(self.query_func):
            raise InvalidParameterError(
                "query_func",
                actual=self.query_func,
                reason="must be synchronous when 'n_workers = 1'",
            )
        elif self.n_workers > 1 and not iscoroutinefunction(self.query_func):
            raise InvalidParameterError(
                "query_func",
                actual=self.query_func,
                reason="must be asynchronous when 'n_workers > 1'",
            )
        if self.fmt not in ["jsonl", "json", "csv"]:
            raise InvalidParameterError(
                "fmt",
                actual=self.fmt,
                reason="must be one of 'jsonl', 'json', and 'csv'",
            )
        if self.logging_mode not in ["all", "failed", "none"]:
            raise InvalidParameterError(
                "logging_mode",
                actual=self.logging_mode,
                reason="must be one of 'all', 'failed', and 'none'",
            )

    def _yield_dataset(
        self, overwrite: bool = False
    ) -> Generator[tuple[int, DataItemType], Any, None]:
        """Yield the indices and data items to be done.

        Yield
        -----
        i : int
            The index of the data item.
        data_item : DataItemType
            The data item.
        """
        source: str | Path
        using_dataset: bool = True

        if not os.path.exists(self.dataset):
            ensure_path(self.dataset)
            source = self.orig_dataset
            using_dataset = False
        else:
            source = self.dataset

        # Load the all data from the best source
        self._all_data: list | pd.DataFrame
        if self.fmt == "jsonl":
            with open(source, "r", encoding="utf-8") as f:
                self._all_data = [json.loads(line) for line in f]
        elif self.fmt == "json":
            with open(source, "r", encoding="utf-8") as f:
                self._all_data = json.load(f)
        else:  # self.fmt == "csv"
            self._all_data = pd.read_csv(source)

        # Yield the indices and corresponding data items
        if using_dataset and not overwrite:
            if self.fmt == "jsonl" or self.fmt == "json":
                for i, data_item in enumerate(self._all_data):
                    if self.response_name not in data_item or pd.isna(
                        data_item[self.response_name]
                    ):
                        yield i, data_item
            else:  # self.format == "csv"
                assert isinstance(
                    self._all_data, pd.DataFrame
                ), "wrong data format; this is most likely an ml3m bug."
                if self.response_name not in self._all_data.columns:
                    # pd.DataFrame index is Hashable | None, must asserting we get int
                    for (  # type: ignore[assignment]
                        i,
                        data_item,
                    ) in self._all_data.iterrows():
                        yield i, data_item
                else:
                    # pd.DataFrame index is Hashable | None, must asserting we get int
                    for i, data_item in self._all_data[  # type: ignore[assignment]
                        self._all_data[self.response_name].isna()
                    ].iterrows():
                        yield i, data_item
        else:
            if self.fmt == "jsonl" or self.fmt == "json":
                for i, data_item in enumerate(self._all_data):
                    yield i, data_item
            else:  # self.format == "csv"
                assert isinstance(
                    self._all_data, pd.DataFrame
                ), "wrong data format; this is most likely an ml3m bug."
                # pd.DataFrame index is Hashable | None, must asserting we get int
                for (  # type: ignore[assignment]
                    i,
                    data_item,
                ) in self._all_data.iterrows():
                    yield i, data_item

    def _update_all_data_list_or_dict(self, i: int, response: str | None) -> None:
        """Update a certain index in the stored data.

        This is only used when ``fmt="jsonl"`` or ``fmt="json"``.

        Parameters
        ----------
        i : int
            The index to update.
        response : str or None
            The response to use for update.
        """
        item = self._all_data[i]
        if isinstance(item, dict):
            item[self.response_name] = response
        elif isinstance(item, list):
            self._all_data[i] = {"data": item, self.response_name: response}
        else:
            # This is unlikely to happen if the code is correct
            raise ValueError(
                f"Each data item must be a list or a dictionary; got '{item}' "
                f"of type '{type(item)}'."
            )  # pragma: no cover

    def generate(self, *, overwrite: bool = False) -> bool:
        """Generate responses and combine with the original dataset.

        Parameters
        ----------
        overwrite : bool, default=False
            Whether to overwrite the responses if some already exist, specified by
            ``response_name``.

        Returns
        -------
        completed : bool
            Whether the task has been completed.
        """
        mlog_path = manage_timed_logs(prefix=type(self).__name__)

        def process_func(
            item: tuple[int, DataItemType], **kwargs
        ) -> (
            tuple[tuple[int, str], list[tuple[Any, Any]], None]
            | tuple[None, None, tuple[Any, Any]]
        ):
            """The sequential processing function."""
            i, data_item = item
            response: str | None = None
            norm_msg: list[tuple[Any, Any]] | None = None
            err: Exception | None = None
            err_trace: str | None = None

            # Handle all exceptions
            try:
                # `process_func` should be called only when using a single worker
                # In that case, `self.query_func` is already validated synchrnous
                formatted_query = self.info_func(data_item)
                response = self.query_func(formatted_query)
                norm_msg = []
                if isinstance(formatted_query, dict):
                    for k, v in formatted_query.items():
                        norm_msg.append((f"[{i}/ QY.{k}]", v))
                elif isinstance(formatted_query, (list, tuple)):
                    for k, v in enumerate(formatted_query):
                        norm_msg.append((f"[{i}/ QY.{k}]", v))
                else:
                    norm_msg.append((f"[{i}/ QY]", formatted_query))
                norm_msg.append((f"[{i}/ RE]", response))
            except Exception as e:
                err, err_trace = e, traceback.format_exc()

            # Write the log on demand
            if (
                self.logging_mode == "failed"
                and response is None
                or self.logging_mode == "all"
            ):
                mlog_item = {
                    "time": str(datetime.now()),
                    "index": i,
                    "response": response,
                    "norm_msg": str(norm_msg),
                    "err_msg": err_trace,
                }
                with open(mlog_path, "a", encoding="utf-8") as f:
                    f.write(json.dumps(mlog_item, ensure_ascii=False) + "\n")

            # Return the information based on success or failure
            if response is not None and norm_msg is not None:
                return (i, response), norm_msg, None
            return None, None, (f"[{i}/ ERR]", f"{type(err).__name__}: {err}")

        async def process_afunc(
            item: tuple[int, DataItemType],
            addtlks: list[asyncio.Lock] | None = None,
            **kwargs,
        ) -> tuple[tuple[int, str], list[tuple[Any, Any]], None] | tuple[
            None, None, tuple[Any, Any]
        ]:
            """The asynchronous processing function."""
            i, data_item = item
            response: str | None = None
            norm_msg: list[tuple[Any, Any]] | None = None
            err: Exception | None = None
            err_trace: str | None = None

            # Handle all exceptions
            try:
                # `process_afunc` should be called only when using multiple workers
                # In that case, `self.query_func` is already validated asynchronous
                formatted_query = self.info_func(data_item)
                response = await self.query_func(formatted_query)  # type: ignore[misc]
                norm_msg = []
                if isinstance(formatted_query, dict):
                    for k, v in formatted_query.items():
                        norm_msg.append((f"[{i}/ QY.{k}]", v))
                elif isinstance(formatted_query, (list, tuple)):
                    for k, v in enumerate(formatted_query):
                        norm_msg.append((f"[{i}/ QY.{k}]", v))
                else:
                    norm_msg.append((f"[{i}/ QY]", formatted_query))
                norm_msg.append((f"[{i}/ RE]", response))
            except Exception as e:
                err, err_trace = e, traceback.format_exc()

            # Write the log on demand
            if (
                self.logging_mode == "failed"
                and response is None
                or self.logging_mode == "all"
            ):
                mlog_item = {
                    "time": str(datetime.now()),
                    "index": i,
                    "response": response,
                    "norm_msg": str(norm_msg),
                    "err_msg": err_trace,
                }
                assert (
                    isinstance(addtlks, list) and len(addtlks) >= 1
                ), "no lock for log in async mode; this is most likely an ml3m bug."
                async with addtlks[0]:
                    with open(mlog_path, "a", encoding="utf-8") as f:
                        f.write(json.dumps(mlog_item, ensure_ascii=False) + "\n")

            # Return the information based on success or failure
            if response is not None and norm_msg is not None:
                return (i, response), norm_msg, None
            return None, None, (f"[{i}/ ERR]", f"{type(err).__name__}: {err}")

        # Activate the asynchronous runner (sequential mode if only one worker)
        runner = AsyncRunner(
            process_func=process_func,
            process_afunc=process_afunc,
            verbose=self.verbose,
        )
        results: list[tuple[int, str]]
        failed: list[tuple[int, DataItemType]]
        results, failed = runner.run(
            items=self._yield_dataset(overwrite=overwrite),
            worker_kwargs=[{} for _ in range(self.n_workers)],
            n_locks=1,
        )
        completed = len(failed) == 0

        # Update the file with the obtained results; all items must be updated,
        # including the failing ones, which should be marked as None
        result_responses = dict(results)
        failed_indices = [item[0] for item in failed]
        if self.fmt == "jsonl" or self.fmt == "json":
            for i, response in result_responses.items():
                self._update_all_data_list_or_dict(i, response)
            for i in failed_indices:
                self._update_all_data_list_or_dict(i, None)
            with open(self.dataset, "w", encoding="utf-8") as f:
                if self.fmt == "jsonl":
                    for data_item in self._all_data:
                        f.write(json.dumps(data_item, ensure_ascii=False) + "\n")
                else:  # self.fmt == "json"
                    json.dump(self._all_data, f, ensure_ascii=False, indent=4)
        else:  # self.fmt == "csv"
            assert isinstance(
                self._all_data, pd.DataFrame
            ), "wrong data format; this is most likely an ml3m bug."
            for i, response in result_responses.items():
                self._all_data.at[i, self.response_name] = response
            for i in failed_indices:
                self._all_data.at[i, self.response_name] = None
            self._all_data.to_csv(self.dataset, index=False)

        # Summarize the save location (and possibly log location)
        print(colored("Dataset can be found at:", COLOR.GREEN))
        print(os.path.abspath(self.dataset))
        if self.logging_mode != "none" and os.path.exists(mlog_path):
            print(colored("Execution log can be found at:", COLOR.GREEN))
            print(os.path.abspath(os.path.abspath(mlog_path)))
        return completed



"""This file is for version-related functionalities."""

import platform
import sys
from importlib.metadata import PackageNotFoundError, version

from ._display import COLOR, colored


def show_versions():
    """Print useful debugging information."""
    from . import __version__

    # Adapted from the scikit-learn implementation
    print()
    welcome_msg = f"Welcome to ml3m {__version__}"
    print(welcome_msg)
    print("=" * len(welcome_msg))

    # Print system related information
    print()
    print(colored("System Information", COLOR.GREEN))
    pyver, (pybuildno, pybuilddt) = platform.python_version(), platform.python_build()
    print(f"Python       {pyver} ({pybuildno}, {pybuilddt})")
    print(f"Compiler     {platform.python_compiler()}")
    print(f"Executable   {sys.executable}")
    print(f"Machine      {platform.platform()}")

    # Print python dependencies
    print()
    print(colored("Python dependencies", COLOR.GREEN))
    packages = ["pip", "setuptools", "numpy", "openai", "pandas", "tqdm"]
    for package in packages:
        try:
            package_ver = version(package)
        except PackageNotFoundError:
            package_ver = None
        print(f"{package:<13}{package_ver}")



from __future__ import annotations

import json
import os
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path


class _OpenAIConfig:
    """OpenAI configuration.

    Parameters
    ----------
    key : str
        The OpenAI API key.
    n_workers : int
        The maximum number of workers to parallelize using this OpenAI API key.
    base : str or None
        The OpenAI API base. ``None`` to use the default base.
    """

    def __init__(self, key: str, n_workers: int, base: str | None = None):
        self.key = key
        self.n_workers = n_workers
        self.base = base or "https://api.openai.com/v1"

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__} <\n    \033[92mkey\033[0m {self.key},\n"
            f"    \033[92mbase\033[0m {self.base},\n    \033[92mn_workers\033[0m "
            f"{self.n_workers},\n>"
        )


def get_openai_config(config_path: str | Path) -> list[_OpenAIConfig]:
    """Get the configurations for OpenAI.

    Parameters
    ----------
    config_path : str or pathlib.Path
        The absolute path to the configuration file.

    Returns
    -------
    openai_configs : list
        The list of OpenAI configuration objects.

    Examples
    --------
    Assume that the configuration file ``.config/openai.json`` looks like the
    following:

    .. code-block::

        [
            {
                "key": "sk-xxx1",
                "base": null,
                "n_workers": 30
            },
            {
                "key": "sk-xxx2",
                "base": "http://127.0.0.1/",
                "n_workers": 5
            }
        ]

    >>> get_openai_config(".config/openai.json")  # doctest: +SKIP
    [_OpenAIConfig <
        key sk-xxx1,
        base https://api.openai.com/v1,
        n_workers 30,
    >, _OpenAIConfig <
        key sk-xxx2,
        base http://127.0.0.1/,
        n_workers 5,
    >]
    """
    abs_config_path = os.path.abspath(config_path)
    with open(abs_config_path, "r", encoding="utf-8") as f:
        configs: list[dict[str, str]] = json.load(f)
    openai_configs = [
        _OpenAIConfig(
            key=config["key"],
            n_workers=int(config["n_workers"]),
            base=config.get("base", None),
        )
        for config in configs
    ]

    # Validate the loaded configurations (not exhaustive)
    if len(openai_configs) == 0:
        raise ValueError("No valid OpenAI configuration found.")
    key_set = {openai_config.key for openai_config in openai_configs}
    if len(key_set) != len(openai_configs):
        raise ValueError("Duplicate OpenAI API keys found in the configuration file.")
    return openai_configs



import asyncio
import copy
import json
import os
import re
from functools import partial

import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from ml3m.base import ResponseGenerator
from ml3m.errors import InvalidParameterError

#######################################################################################
#                                                                                     #
#                                        DATA                                         #
#                                                                                     #
#######################################################################################


orig_dataset_2 = [
    {
        "instruction": "What is the capital of China?",
        "output": "The capital of China is Beijing.",
    },
    {
        "instruction": "What is the capital of France?",
        "output": "The capital of France is Paris.",
    },
]

orig_dataset_l2 = [
    ["What is the capital of China?", "The capital of China is Beijing."],
    ["What is the capital of France?", "The capital of France is Paris."],
]


@pytest.fixture(scope="module")
def prepare(request, storage):
    """Make a temporary storage and clear it towards the end."""
    paths = {}

    # Make files for `orig_dataset_2`
    for fmt in ["jsonl", "json", "csv"]:
        dataset = os.path.join(
            storage, f"orig_dataset_2__{request.keywords.node.name}.{fmt}"
        )
        if fmt == "jsonl":
            with open(dataset, "w", encoding="utf-8") as f:
                for item in orig_dataset_2:
                    f.write(json.dumps(item, ensure_ascii=False) + "\n")
        elif fmt == "json":
            with open(dataset, "w", encoding="utf-8") as f:
                json.dump(orig_dataset_2, f, ensure_ascii=False, indent=4)
        else:  # fmt == "csv"
            df = pd.DataFrame(orig_dataset_2)
            df.to_csv(dataset, index=False)
        paths[f"orig_dataset_2__{fmt}"] = dataset

    # Make files for `orig_dataset_l2`
    for fmt in ["jsonl", "json"]:
        dataset = os.path.join(
            storage, f"orig_dataset_l2__{request.keywords.node.name}.{fmt}"
        )
        with open(dataset, "w", encoding="utf-8") as f:
            if fmt == "jsonl":
                for item in orig_dataset_l2:
                    f.write(json.dumps(item, ensure_ascii=False) + "\n")
            else:  # fmt == "json":
                json.dump(orig_dataset_l2, f, ensure_ascii=False, indent=4)
        paths[f"orig_dataset_l2__{fmt}"] = dataset

    return paths


#######################################################################################
#                                                                                     #
#                                  PREPARATION WORK                                   #
#                                                                                     #
#######################################################################################


def query_func_fixed(query, response="", mode="normal"):
    """Return a fixed response.

    This can normally pass all items, fail all items, fail a certain item based on the
    instruction field, or fail a certain item based on the a certain index (when each
    item is a list).
    """
    if mode == "normal" or mode.startswith("err_on_"):
        if mode.startswith("err_on_instruction_") and query["instruction"] == mode[19:]:
            raise ValueError
        mat = re.match(r"err_on_index\.(\d+)_(.+)", mode, re.DOTALL)
        if mat is not None:
            if isinstance(query, list) and query[int(mat.group(1))] == mat.group(2):
                raise ValueError
            if (
                isinstance(query, dict)
                and "data" in query
                and isinstance(query["data"], list)
                and query["data"][int(mat.group(1))] == mat.group(2)
            ):
                raise ValueError
        return response
    elif mode == "all_err":
        raise ValueError


async def query_afunc_fixed(query, response, mode="normal"):
    """Return a fixed response.

    This can normally pass all items, fail all items, fail a certain item based on the
    instruction field, or fail a certain item based on the a certain index (when each
    item is a list).
    """
    await asyncio.sleep(0.01)
    return query_func_fixed(query, response, mode=mode)


#######################################################################################
#                                                                                     #
#                                  TESTS START HERE                                   #
#                                                                                     #
#######################################################################################


class TestResponseGenerator:
    """Testing ml3m.base.ResponseGenerator."""

    @pytest.mark.parametrize("fmt", ["jsonl", "json", "csv"])
    @pytest.mark.parametrize(
        "n_workers,query_func",
        [
            (1, partial(query_func_fixed, response="I don't know.")),
            (3, partial(query_afunc_fixed, response="I don't know.")),
        ],
    )
    @pytest.mark.parametrize("response_name", ["response", "model_response"])
    @pytest.mark.parametrize("logging_mode", ["none", "all", "failed"])
    @pytest.mark.parametrize("verbose", [-1, 0, 1, 2])
    def test_response_generator_result_versus_written(
        self,
        query_func,
        response_name,
        fmt,
        n_workers,
        logging_mode,
        verbose,
        storage,
        prepare,
        request,
    ):
        """Test that generator._all_data and the written dataset are the same.

        This serves as a basis so that the other tests do not need to I/O to check the
        results but instead directly read the attribute.
        """
        orig_dataset = prepare[f"orig_dataset_2__{fmt}"]
        dataset = os.path.join(storage, f"dataset__{request.keywords.node.name}.{fmt}")

        generator = ResponseGenerator(
            orig_dataset=orig_dataset,
            dataset=dataset,
            info_func=lambda x: x,
            query_func=query_func,
            response_name=response_name,
            fmt=fmt,
            n_workers=n_workers,
            logging_mode=logging_mode,
            verbose=verbose,
        )
        completed = generator.generate()
        assert completed

        if fmt == "jsonl":
            with open(dataset, "r", encoding="utf-8") as f:
                data_saved = [json.loads(line) for line in f]
            assert data_saved == generator._all_data

        elif fmt == "json":
            with open(dataset, "r", encoding="utf-8") as f:
                data_saved = json.load(f)
            assert data_saved == generator._all_data

        else:  # fmt == "csv"
            data_saved = pd.read_csv(dataset)
            assert_frame_equal(data_saved, generator._all_data)

    @pytest.mark.parametrize("fmt", ["jsonl", "json"])
    @pytest.mark.parametrize(
        "n_workers,query_func",
        [
            (1, partial(query_func_fixed, response="I don't know.")),
            (3, partial(query_afunc_fixed, response="I don't know.")),
        ],
    )
    @pytest.mark.parametrize("response_name", ["response", "model_response"])
    @pytest.mark.parametrize("logging_mode", ["none", "all", "failed"])
    @pytest.mark.parametrize("verbose", [-1, 0, 1, 2])
    def test_response_generator_result_versus_written_list(
        self,
        query_func,
        response_name,
        fmt,
        n_workers,
        logging_mode,
        verbose,
        storage,
        prepare,
        request,
    ):
        """Test that generator._all_data and the written dataset are the same.

        This serves as a basis so that the other tests do not need to I/O to check the
        results but instead directly read the attribute.

        This is the special case where each data item is a list.
        """
        orig_dataset = prepare[f"orig_dataset_l2__{fmt}"]
        dataset = os.path.join(storage, f"dataset__{request.keywords.node.name}.{fmt}")

        generator = ResponseGenerator(
            orig_dataset=orig_dataset,
            dataset=dataset,
            info_func=lambda x: x,
            query_func=query_func,
            response_name=response_name,
            fmt=fmt,
            n_workers=n_workers,
            logging_mode=logging_mode,
            verbose=verbose,
        )
        completed = generator.generate()
        assert completed

        with open(dataset, "r", encoding="utf-8") as f:
            if fmt == "jsonl":
                data_saved = [json.loads(line) for line in f]
            else:  # fmt == "json"
                data_saved = json.load(f)
        assert data_saved == generator._all_data

    @pytest.mark.parametrize("fmt", ["jsonl", "json", "csv"])
    @pytest.mark.parametrize(
        "n_workers,query_func",
        [
            (1, partial(query_func_fixed, response="I don't know.")),
            (3, partial(query_afunc_fixed, response="I don't know.")),
        ],
    )
    def test_response_generator_generate_basics(
        self, fmt, n_workers, query_func, storage, prepare, request
    ):
        """Test the basic generator functionalities.

        Fail all data items
        -> Pass one of the data items
        -> Pass all data items
        -> Generate again (should make no change)
        -> Fail all data items with the new response name (should make no change)
        -> Pass one of the data items with the new response name
        -> Generate on the old response name again (should make no change)
        -> Pass all data items with the new response name
        -> Generate again (should make no change)
        """
        orig_dataset = prepare[f"orig_dataset_2__{fmt}"]
        dataset = os.path.join(storage, f"dataset__{request.keywords.node.name}.{fmt}")

        def make_generator_by_mode(mode, response_name):
            """Convenient function for making a generator based on mode."""
            return ResponseGenerator(
                orig_dataset=orig_dataset,
                dataset=dataset,
                info_func=lambda x: x,
                query_func=partial(query_func, mode=mode),
                response_name=response_name,
                fmt=fmt,
                n_workers=n_workers,
            )

        # This should pass none of the data items
        generator = make_generator_by_mode("all_err", "response")
        completed = generator.generate()
        assert not completed

        results = generator._all_data
        if fmt == "json" or fmt == "jsonl":
            expected = copy.deepcopy(orig_dataset_2)
            expected[0]["response"] = None
            expected[1]["response"] = None
            assert results == expected
        else:  # fmt == "csv"
            expected_df = pd.DataFrame(orig_dataset_2)
            expected_df["response"] = None
            assert_frame_equal(expected_df, results, check_dtype=False)

        # This should pass the second data item but fail the firrst
        item_0_instruction = orig_dataset_2[0]["instruction"]
        generator = make_generator_by_mode(
            f"err_on_instruction_{item_0_instruction}", "response"
        )
        completed = generator.generate()
        assert not completed

        results = generator._all_data
        if fmt == "json" or fmt == "jsonl":
            expected[1]["response"] = "I don't know."
            assert results == expected
        else:  # fmt == "csv"
            expected_df.at[1, "response"] = "I don't know."
            assert_frame_equal(expected_df, results, check_dtype=False)

        # This should pass all of the data items
        generator = make_generator_by_mode("normal", "response")
        completed = generator.generate()
        assert completed

        results = generator._all_data
        if fmt == "json" or fmt == "jsonl":
            expected[0]["response"] = "I don't know."
            assert results == expected
        else:  # fmt == "csv"
            expected_df.at[0, "response"] = "I don't know."
            assert_frame_equal(expected_df, results, check_dtype=False)

        # This should not modify the generated dataset
        generator.generate()
        completed = generator.generate()
        assert completed

        results = generator._all_data
        if fmt == "json" or fmt == "jsonl":
            assert results == expected
        else:  # fmt == "csv"
            assert_frame_equal(expected_df, results, check_dtype=False)

        # This should pass none of the data items with the new response name
        generator2 = make_generator_by_mode("all_err", "new_response")
        completed = generator2.generate()
        assert not completed

        results = generator2._all_data
        if fmt == "json" or fmt == "jsonl":
            expected[0]["new_response"] = None
            expected[1]["new_response"] = None
            assert results == expected
        else:  # fmt == "csv"
            expected_df["new_response"] = None
            assert_frame_equal(expected_df, results, check_dtype=False)

        # This should pass the second item but fail the first with the new response
        # name
        generator2 = make_generator_by_mode(
            f"err_on_instruction_{item_0_instruction}", "new_response"
        )
        completed = generator2.generate()
        assert not completed

        results = generator2._all_data
        if fmt == "json" or fmt == "jsonl":
            expected[1]["new_response"] = "I don't know."
            assert results == expected
        else:  # fmt == "csv"
            expected_df.at[1, "new_response"] = "I don't know."
            assert_frame_equal(expected_df, results, check_dtype=False)

        # This should not modify the generated dataset since it is using the old
        # response name
        completed = generator.generate()
        assert completed

        results = generator._all_data
        if fmt == "json" or fmt == "jsonl":
            assert results == expected
        else:  # fmt == "csv"
            assert_frame_equal(expected_df, results, check_dtype=False)

        # This should pass all of the data items with the new response name
        generator2 = make_generator_by_mode("normal", "new_response")
        completed = generator2.generate()
        assert completed

        results = generator2._all_data
        if fmt == "json" or fmt == "jsonl":
            expected[0]["new_response"] = "I don't know."
            assert results == expected
        else:  # fmt == "csv"
            expected_df.at[0, "new_response"] = "I don't know."
            assert_frame_equal(expected_df, results, check_dtype=False)

        # This should not modify the generated dataset
        completed = generator2.generate()
        assert completed

        results = generator2._all_data
        if fmt == "json" or fmt == "jsonl":
            assert results == expected
        else:  # fmt == "csv"
            assert_frame_equal(expected_df, results, check_dtype=False)

    @pytest.mark.parametrize("fmt", ["jsonl", "json"])
    @pytest.mark.parametrize(
        "n_workers,query_func",
        [
            (1, partial(query_func_fixed, response="I don't know.")),
            (3, partial(query_afunc_fixed, response="I don't know.")),
        ],
    )
    def test_response_generator_generate_basics_list(
        self, fmt, n_workers, query_func, storage, prepare, request
    ):
        """Test the basic generator functionalities.

        Fail all data items
        -> Pass one of the data items
        -> Pass all data items
        -> Generate again (should make no change)
        -> Fail all data items with the new response name (should make no change)
        -> Pass one of the data items with the new response name
        -> Generate on the old response name again (should make no change)
        -> Pass all data items with the new response name
        -> Generate again (should make no change)

        This is the special case where each data item is a list.
        """
        orig_dataset = prepare[f"orig_dataset_l2__{fmt}"]
        dataset = os.path.join(storage, f"dataset__{request.keywords.node.name}.{fmt}")

        def make_generator_by_mode(mode, response_name):
            """Convenient function for making a generator based on mode."""
            return ResponseGenerator(
                orig_dataset=orig_dataset,
                dataset=dataset,
                info_func=lambda x: x,
                query_func=partial(query_func, mode=mode),
                response_name=response_name,
                fmt=fmt,
                n_workers=n_workers,
            )

        # This should pass none of the data items
        generator = make_generator_by_mode("all_err", "response")
        completed = generator.generate()
        assert not completed

        results = generator._all_data
        expected = [
            {"data": orig_dataset_l2[0], "response": None},
            {"data": orig_dataset_l2[1], "response": None},
        ]
        assert results == expected

        # This should pass the second data item but fail the firrst
        item_0_instruction = orig_dataset_l2[0][0]
        generator = make_generator_by_mode(
            f"err_on_index.0_{item_0_instruction}", "response"
        )
        completed = generator.generate()
        assert not completed

        results = generator._all_data
        expected[1]["response"] = "I don't know."
        assert results == expected

        # This should pass all of the data items
        generator = make_generator_by_mode("normal", "response")
        completed = generator.generate()
        assert completed

        results = generator._all_data
        expected[0]["response"] = "I don't know."
        assert results == expected

        # This should not modify the generated dataset
        generator.generate()
        completed = generator.generate()
        assert completed

        results = generator._all_data
        assert results == expected

        # This should pass none of the data items with the new response name
        generator2 = make_generator_by_mode("all_err", "new_response")
        completed = generator2.generate()
        assert not completed

        results = generator2._all_data
        expected[0]["new_response"] = None
        expected[1]["new_response"] = None
        assert results == expected

        # This should pass the second item but fail the first with the new response
        # name
        generator2 = make_generator_by_mode(
            f"err_on_index.0_{item_0_instruction}", "new_response"
        )
        completed = generator2.generate()
        assert not completed

        results = generator2._all_data
        expected[1]["new_response"] = "I don't know."
        assert results == expected

        # This should not modify the generated dataset since it is using the old
        # response name
        completed = generator.generate()
        assert completed

        results = generator._all_data
        assert results == expected

        # This should pass all of the data items with the new response name
        generator2 = make_generator_by_mode("normal", "new_response")
        completed = generator2.generate()
        assert completed

        results = generator2._all_data
        expected[0]["new_response"] = "I don't know."
        assert results == expected

        # This should not modify the generated dataset
        completed = generator2.generate()
        assert completed

        results = generator2._all_data
        assert results == expected

    @pytest.mark.parametrize("fmt", ["jsonl", "json", "csv"])
    @pytest.mark.parametrize(
        "n_workers,query_func",
        [(1, query_func_fixed), (3, query_afunc_fixed)],
    )
    def test_response_generator_overwrite(
        self, fmt, n_workers, query_func, storage, prepare, request
    ):
        """Test overwrite parameter of evaluate.

        Pass all data items
        -> Overwrite and fail all data items
        -> Overwrite and pass all data items
        -> Overwrite and pass all data items with the new response name
        -> Overwrite and fail all data items with the new response name
        -> Pass all data items with the new response name
        -> Overwrite and pass all data items with the new response name
        """
        orig_dataset = prepare[f"orig_dataset_2__{fmt}"]
        dataset = os.path.join(storage, f"dataset__{request.keywords.node.name}.{fmt}")

        def make_generator_by_mode(mode, response_name, response):
            """Convenient function for making a generator based on mode."""
            return ResponseGenerator(
                orig_dataset=orig_dataset,
                dataset=dataset,
                info_func=lambda x: x,
                query_func=partial(query_func, response=response, mode=mode),
                response_name=response_name,
                fmt=fmt,
                n_workers=n_workers,
            )

        # Preparation: Pass all data items
        generator = make_generator_by_mode("normal", "response", "a")
        generator.generate()

        # Overwrite and fail all data items
        generator = make_generator_by_mode("all_err", "response", "b")
        completed = generator.generate(overwrite=True)
        assert not completed

        results = generator._all_data
        if fmt == "json" or fmt == "jsonl":
            expected = copy.deepcopy(orig_dataset_2)
            expected[0]["response"] = None
            expected[1]["response"] = None
            assert results == expected
        else:  # fmt == "csv"
            expected_df = pd.DataFrame(orig_dataset_2)
            expected_df["response"] = None
            assert_frame_equal(expected_df, results, check_dtype=False)

        # Overwrite and pass all data items
        generator = make_generator_by_mode("normal", "response", "b")
        completed = generator.generate(overwrite=True)
        assert completed

        results = generator._all_data
        if fmt == "json" or fmt == "jsonl":
            expected[0]["response"] = "b"
            expected[1]["response"] = "b"
            assert results == expected
        else:  # fmt == "csv"
            expected_df["response"] = "b"
            assert_frame_equal(expected_df, results, check_dtype=False)

        # Overwrite and pass all data items with the new response name
        generator2 = make_generator_by_mode("normal", "new_response", "c")
        completed = generator2.generate(overwrite=True)
        assert completed

        results = generator2._all_data
        if fmt == "json" or fmt == "jsonl":
            expected[0]["new_response"] = "c"
            expected[1]["new_response"] = "c"
            assert results == expected
        else:  # fmt == "csv"
            expected_df["new_response"] = "c"
            assert_frame_equal(expected_df, results, check_dtype=False)

        # Overwrite and fail all data items with the new response name
        generator2 = make_generator_by_mode("all_err", "new_response", "d")
        completed = generator2.generate(overwrite=True)
        assert not completed

        results = generator2._all_data
        if fmt == "json" or fmt == "jsonl":
            expected[0]["new_response"] = None
            expected[1]["new_response"] = None
            assert results == expected
        else:  # fmt == "csv"
            expected_df["new_response"] = None
            assert_frame_equal(expected_df, results, check_dtype=False)

        # Preparation: Pass all data items with the new response name
        generator2 = make_generator_by_mode("normal", "new_response", "d")
        generator2.generate()

        # Overwrite and pass all data items with the new response name
        generator2 = make_generator_by_mode("normal", "new_response", "e")
        completed = generator2.generate(overwrite=True)
        assert completed

        results = generator2._all_data
        if fmt == "json" or fmt == "jsonl":
            expected[0]["new_response"] = "e"
            expected[1]["new_response"] = "e"
            assert results == expected
        else:  # fmt == "csv"
            expected_df["new_response"] = "e"
            assert_frame_equal(expected_df, results, check_dtype=False)

    @pytest.mark.parametrize("fmt", ["jsonl", "json"])
    @pytest.mark.parametrize(
        "n_workers,query_func",
        [(1, query_func_fixed), (3, query_afunc_fixed)],
    )
    def test_response_generator_overwrite_list(
        self, fmt, n_workers, query_func, storage, prepare, request
    ):
        """Test overwrite parameter of evaluate.

        Pass all data items
        -> Overwrite and fail all data items
        -> Overwrite and pass all data items
        -> Overwrite and pass all data items with the new response name
        -> Overwrite and fail all data items with the new response name
        -> Pass all data items with the new response name
        -> Overwrite and pass all data items with the new response name

        This is the special case where each data item is a list.
        """
        orig_dataset = prepare[f"orig_dataset_l2__{fmt}"]
        dataset = os.path.join(storage, f"dataset__{request.keywords.node.name}.{fmt}")

        def make_generator_by_mode(mode, response_name, response):
            """Convenient function for making a generator based on mode."""
            return ResponseGenerator(
                orig_dataset=orig_dataset,
                dataset=dataset,
                info_func=lambda x: x,
                query_func=partial(query_func, response=response, mode=mode),
                response_name=response_name,
                fmt=fmt,
                n_workers=n_workers,
            )

        # Preparation: Pass all data items
        generator = make_generator_by_mode("normal", "response", "a")
        generator.generate()

        # Overwrite and fail all data items
        generator = make_generator_by_mode("all_err", "response", "b")
        completed = generator.generate(overwrite=True)
        assert not completed

        results = generator._all_data
        expected = [
            {"data": orig_dataset_l2[0], "response": None},
            {"data": orig_dataset_l2[1], "response": None},
        ]
        assert results == expected

        # Overwrite and pass all data items
        generator = make_generator_by_mode("normal", "response", "b")
        completed = generator.generate(overwrite=True)
        assert completed

        results = generator._all_data
        expected[0]["response"] = "b"
        expected[1]["response"] = "b"
        assert results == expected

        # Overwrite and pass all data items with the new response name
        generator2 = make_generator_by_mode("normal", "new_response", "c")
        completed = generator2.generate(overwrite=True)
        assert completed

        results = generator2._all_data
        expected[0]["new_response"] = "c"
        expected[1]["new_response"] = "c"
        assert results == expected

        # Overwrite and fail all data items with the new response name
        generator2 = make_generator_by_mode("all_err", "new_response", "d")
        completed = generator2.generate(overwrite=True)
        assert not completed

        results = generator2._all_data
        expected[0]["new_response"] = None
        expected[1]["new_response"] = None
        assert results == expected

        # Preparation: Pass all data items with the new response name
        generator2 = make_generator_by_mode("normal", "new_response", "d")
        generator2.generate()

        # Overwrite and pass all data items with the new response name
        generator2 = make_generator_by_mode("normal", "new_response", "e")
        completed = generator2.generate(overwrite=True)
        assert completed

        results = generator2._all_data
        expected[0]["new_response"] = "e"
        expected[1]["new_response"] = "e"
        assert results == expected

    def test_response_generator_invalid_init(self, storage, prepare, request):
        """Test invalid initialization."""
        orig_dataset = prepare["orig_dataset_2__jsonl"]
        dataset = os.path.join(storage, f"dataset__{request.keywords.node.name}.jsonl")

        # Test invalid n_workers
        for n_workers in [0, 2.4]:
            with pytest.raises(InvalidParameterError):
                ResponseGenerator(
                    orig_dataset=orig_dataset,
                    dataset=dataset,
                    info_func=lambda x: x,
                    query_func=query_func_fixed,
                    response_name="response",
                    n_workers=n_workers,
                )

        # Test invalid info_func
        for info_func in [None, "func"]:
            with pytest.raises(InvalidParameterError):
                ResponseGenerator(
                    orig_dataset=orig_dataset,
                    dataset=dataset,
                    info_func=info_func,
                    query_func=query_func_fixed,
                    response_name="response",
                )

        # Test invalid query_func
        for query_func in [None, "func"]:
            with pytest.raises(InvalidParameterError):
                ResponseGenerator(
                    orig_dataset=orig_dataset,
                    dataset=dataset,
                    info_func=lambda x: x,
                    query_func=query_func,
                    response_name="response",
                )

        # Test incompatible n_workers with query_func
        for n_workers, query_func in zip(
            [1, 3],
            [query_afunc_fixed, query_func_fixed],
        ):
            with pytest.raises(InvalidParameterError):
                ResponseGenerator(
                    orig_dataset=orig_dataset,
                    dataset=dataset,
                    info_func=lambda x: x,
                    query_func=query_func,
                    response_name="response",
                    n_workers=n_workers,
                )

        # Test invalid fmt
        for fmt in [".csv", "txt"]:
            with pytest.raises(InvalidParameterError):
                ResponseGenerator(
                    orig_dataset=orig_dataset,
                    dataset=dataset,
                    info_func=lambda x: x,
                    query_func=query_func_fixed,
                    response_name="response",
                    fmt=fmt,
                )

        # Test invalid logging_mode
        for logging_mode in ["any", "succeeded"]:
            with pytest.raises(InvalidParameterError):
                ResponseGenerator(
                    orig_dataset=orig_dataset,
                    dataset=dataset,
                    info_func=lambda x: x,
                    query_func=query_func_fixed,
                    response_name="response",
                    logging_mode=logging_mode,
                )



"""This file is for path-related functionalities."""


from __future__ import annotations

import os
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path


def ensure_path(path: str | Path, is_directory=False) -> str:
    """Create medium directories ensure path is safe for creation.

    Parameters
    ----------
    path : str or pathlib.Path
        The path to ensure creation safety of.
    is_directory : bool, default=False
        Whether ``path`` is a directory instead of a file.

    Returns
    -------
    directories : str
        The path to the directory of ``path`` if ``path`` is a file. Otherwise this is
        ``path`` itself. In both cases, this would be a formatted absolute path.
    """
    directories = path if is_directory else os.path.split(path)[0]
    if not os.path.exists(directories):
        os.makedirs(directories)
    return os.path.abspath(directories)


def validate_path(path: str | Path) -> None:
    """Validate a path.

    Parameters
    ----------
    path : str or pathlib.Path
        The path to validate.
    """
    if not os.path.exists(path):
        raise ValueError(f"Path '{os.path.abspath(path)}' does not exist.")



from __future__ import annotations

import asyncio
import json
import os
import traceback
from datetime import datetime
from numbers import Real
from typing import TYPE_CHECKING, Any, Generator

import openai
import pandas as pd

from .._async import AsyncRunner
from .._display import COLOR, colored, format_data_item
from .._logging import manage_timed_logs
from .._paths import ensure_path, validate_path
from ..errors import InvalidParameterError, ScoringError
from ..utils.openai import get_openai_config

if TYPE_CHECKING:
    from pathlib import Path

    from .._typing import AggregateMethod, DataItemType, DatasetFormat, LoggingMode


class BaseEvaluator:
    """Base evaluator class.

    .. note::
        This class is meant to be subclassed. The methods that must be overridden
        include:

        - :meth:`BaseEvaluator._aget_score` (if to be used with multiple workers)
        - :meth:`BaseEvaluator._get_score` (if to be used with a single worker)

    Parameters
    ----------
    dataset : str or pathlib.Path
        The absolute path to the evaluation dataset.
    save_path : str or pathlib.Path
        The absolute path to the save location. This path may or may not exist, and if
        it exists, its file contents will be treated as a (partially) written result.
        Whether to overwrite the existing results or to build on them depend on
        ``overwrite`` when using the :meth:`BaseEvaluator.evaluate` method.
    subjects : list
        The subjects to evaluate. This should *strictly* correspond to how to scores
        are obtained in :meth:`BaseEvaluator._aget_score`. If the score is obtained as
        a single real value, ``subjects`` must be a list of one element and that
        element will be used as the name of that score. If the score(s) are obtained as
        a dictionary of subject-value pairs, all items in ``subjects`` must appear as
        the keys of the obtained dictionary. Any additional key will be discarded; any
        missing key will be treated as an error.
    fmt : {"jsonl", "json", "csv"}, default="jsonl"
        The format of ``dataset``.
    workers : int or list of dict, default=1
        If ``workers`` is an integer, it will be considered the number of workers. If
        specified only one worker, the dataset will be processed sequentially, and
        otherwise it will be asynchronously parallelized. If ``workers`` is a list of
        dictionaries, the length of this list will be considered the number of workers.
        Each dictionary should be the additional keyword arguments passed to
        :meth:`BaseEvaluator._aget_score`. Note that if ``workers`` is an integer, no
        additional keyword arguments will be passed.
    n_iter : int, default=1
        The number of iterations for each data item. This is commonly used when one
        round of scoring is not convincing enough and the final scoring should be some
        statistics of multiple rounds of scoring.
    agg_method : {"mean", "sum", "min", "max", "mode"}, default=None
        The aggregate method to use on multiple rounds of scoring. Ignored when
        ``n_iter=1``, and otherwise this must not be ``None``.
    logging_mode : {"all", "failed", "none"}, default="all"
        The logging mode, whether to save the logs of all items, or only of failed
        items, or save no log.
    verbose : int, default=0
        The verbosity level of the processing. For negative levels, only a progress bar
        will be displayed. For level 0, the errored items will also be displayed. For
        positive levels, the all items will be displayed, and the verbosity level
        determines the number of lines to display for the message of each item.
    """

    def __init__(
        self,
        dataset: str | Path,
        save_path: str | Path,
        subjects: list,
        *,
        fmt: DatasetFormat = "jsonl",
        workers: int | list[dict] = 1,
        n_iter: int = 1,
        agg_method: AggregateMethod | None = None,
        logging_mode: LoggingMode = "all",
        verbose: int = 0,
    ) -> None:
        self.dataset = dataset
        self.save_path = save_path
        self.subjects = subjects
        self.fmt = fmt
        self.workers = workers
        self.n_iter = n_iter
        self.agg_method = agg_method
        self.logging_mode = logging_mode
        self.verbose = verbose

        # Validate the arguments
        validate_path(self.dataset)
        if not isinstance(self.subjects, list):
            raise InvalidParameterError(
                "subjects", actual=self.subjects, reason="must be a list"
            )
        if len(self.subjects) == 0:
            raise InvalidParameterError(
                "subjects", actual=self.subjects, reason="must not be empty"
            )
        if "i" in subjects:
            raise InvalidParameterError(
                "subjects",
                actual=self.subjects,
                reason="'i' is a reserved key for indexing",
            )
        if self.fmt not in ["jsonl", "json", "csv"]:
            raise InvalidParameterError(
                "fmt",
                actual=self.fmt,
                reason="must be one of 'jsonl', 'json', and 'csv'",
            )
        if self.logging_mode not in ["all", "failed", "none"]:
            raise InvalidParameterError(
                "logging_mode",
                actual=self.logging_mode,
                reason="must be one of 'all', 'failed', and 'none'",
            )
        if not isinstance(self.n_iter, int) or self.n_iter < 1:
            raise InvalidParameterError(
                "n_iter", actual=self.n_iter, reason="must be an integer >= 1"
            )
        if n_iter > 1 and self.agg_method not in ["mean", "sum", "min", "max", "mode"]:
            raise InvalidParameterError(
                "agg_method",
                actual=self.agg_method,
                reason=(
                    "when 'n_iter > 1', must be one of 'mean', 'sum', 'min', 'max', "
                    "and 'mode'"
                ),
            )

        # Load the keyword arguments for workers
        self._worker_kwargs: list[dict]
        if isinstance(self.workers, int):
            if self.workers < 1:
                raise InvalidParameterError(
                    "workers",
                    actual=self.workers,
                    reason="if given as an integer, must be >= 1",
                )
            self._worker_kwargs = [{} for _ in range(self.workers)]
        elif isinstance(self.workers, list):
            if any(not isinstance(worker, dict) for worker in self.workers):
                raise InvalidParameterError(
                    "workers",
                    actual=self.workers,
                    reason="if given as a list, each element must be a dict",
                )
            self._worker_kwargs = self.workers
        else:
            raise InvalidParameterError(
                "workers",
                actual=self.workers,
                reason="must be an integer >= 1 or a list of dict",
            )

    def _get_score(self, data_item: DataItemType, **kwargs) -> Real | dict[Any, Real]:
        """Evaluate a data item and obtain its score(s).

        :meta public:

        .. note::
            This method is not implemented and must be overridden in subclasses.

        Parameters
        ----------
        data_item : DataItemType
            The data item.
        kwargs
            The additional keyword arguments.

        Returns
        -------
        scores : real or dict
            The evaluated scores, either a single score or a dictionary of subject-
            score pairs.

        Notes
        -----
        Note that if there are mutliple scores obtained, i.e., returning a dictionary,
        remember that the "i" key cannot be included since it is reserved for indexing.

        Moreover, it is recommended *not* to catch the exceptions that cause the
        processing of a data item to fail, since otherwise
        :meth:`BaseEvaluator.evaluate` will not realize that the data item errors out.
        """
        raise NotImplementedError

    async def _aget_score(
        self, data_item: DataItemType, **kwargs
    ) -> Real | dict[Any, Real]:
        """Evaluate a data item and obtain its score(s).

        :meta public:

        This should be the asynchronous version of :meth:`BaseEvaluator._get_score`.
        See :meth:`BaseEvaluator._get_score` for details. If the obtaining process of
        the scores does not involving anything asynchrounous, this can simply be
        overridden as follows:

        .. code-block:: python

            async def _aget_score(self, data_item, **kwargs):
                return self._get_score(data_item, **kwargs)


        However, note that this will cause only one worker to be actually doing all
        the tasks because it will not give away the control of the event loop.

        .. note::
            This method is not implemented and must be overridden in subclasses.
            Moreover, this method must be defined as asynchronous.
        """
        raise NotImplementedError

    def _yield_dataset(
        self, overwrite: bool = False
    ) -> Generator[tuple[int, int, DataItemType], Any, None]:
        """Yield the indices and data items to be done.

        Parameters
        ----------
        overwrite : bool, default=False
            If ``True``, all existing indices will be ignored.

        Yield
        -----
        i : int
            The index of the data item.
        it : int
            The iteration id of the data item, should be range(0, self.n_iter).
        data_item : DataItemType
            The data item.
        """
        df: pd.DataFrame
        existing_indices: list = []
        if not os.path.exists(self.save_path):
            ensure_path(self.save_path)
            df = pd.DataFrame(columns=["i"], dtype=pd.Int64Dtype)
        else:
            df = pd.read_csv(self.save_path)
            if all(subject in df.columns for subject in self.subjects):
                if not overwrite:
                    existing_indices = list(
                        df[
                            ~df[list(self.subjects)]  # type: ignore[type-var]
                            .isna()
                            .any(axis=1)
                        ].loc[:, "i"]
                    )
        self._result_df = df.set_index("i")

        # Yield the indices and corresponding data items
        if self.fmt == "jsonl":
            with open(self.dataset, "r", encoding="utf-8") as f:
                for i, line in enumerate(f):
                    if i not in existing_indices:
                        for it in range(self.n_iter):
                            yield i, it, json.loads(line)
        elif self.fmt == "json":
            with open(self.dataset, "r", encoding="utf-8") as f:
                all_data = json.load(f)
            for i, item in enumerate(all_data):
                if i not in existing_indices:
                    for it in range(self.n_iter):
                        yield i, it, item
        else:  # self.fmt == "csv"
            all_data = pd.read_csv(self.dataset)
            for i, row in all_data.iterrows():
                if i not in existing_indices:
                    for it in range(self.n_iter):
                        yield i, it, row

    def _check_scores(self, scores: Real | dict[Any, Real]) -> dict[Any, Real]:
        """Check and format the scores.

        Parameters
        ----------
        scores : real or dict
            The evaluation scores of a data item, either a single real number of a
            dictionary of subject-score pairs.

        Returns
        -------
        eval_scores : dict
            If ``scores`` is a single real number, this returns ``{"scores": scores}``.
            Otherwise, this returns ``scores`` itself.
        """
        # mypy not working with numbers.Real
        if isinstance(scores, Real) and not pd.isna(
            scores
        ):  # type: ignore[call-overload]
            if len(self.subjects) == 1:
                return {self.subjects[0]: scores}
            raise ScoringError("Got a single score but multiple subjects are assigned.")
        elif isinstance(scores, dict):
            # mypy not working with numbers.Real
            bad_item = next(
                (
                    (subject, score)
                    for subject, score in scores.items()
                    if (
                        not isinstance(score, Real)
                        or pd.isna(score)  # type: ignore[call-overload]
                    )
                ),
                None,
            )
            if bad_item is not None:
                raise ScoringError(
                    "The scores must be either a real number or a dictionary with "
                    f"real values; got a dictionary but there exists "
                    f"'{bad_item[0]}: {bad_item[1]}' of type '{type(bad_item[1])}'."
                )
            else:
                missing_keys = set(self.subjects).difference(set(scores))
                if missing_keys:
                    missing_keys_repr = ", ".join(map(repr, missing_keys))
                    raise ScoringError(f"Missing keys: {missing_keys_repr}.")
                return {
                    subject: score
                    for subject, score in scores.items()
                    if subject in self.subjects
                }
        else:
            raise ScoringError(
                "The scores must be either a real number or a dictionary with real "
                f"values; got '{scores}' of type '{type(scores)}' instead."
            )

    def evaluate(self, *, overwrite: bool = False) -> bool:
        """Evaluate the specified dataset.

        Parameters
        ----------
        overwrite : bool, default=False
            Whether to overwrite the data in ``save_path``. If ``False``, the
            evaluation will be built upon existing data in ``save_path``, otherwise
            all data will be evaluated are existing data will be overwritten.

        Returns
        -------
        completed : bool
            Whether the task has been completed.
        """
        mlog_path = manage_timed_logs(type(self).__name__)

        def process_func(
            item: tuple[int, int, DataItemType], **kwargs
        ) -> (
            tuple[tuple[int, int, dict[Any, Real]], list[tuple[Any, Any]], None]
            | tuple[None, None, tuple[Any, Any]]
        ):
            """The sequential processing function."""
            i, it, data_item = item
            prefix = f"{i}" if self.n_iter == 1 else f"{i}.{it}"
            eval_scores: dict[Any, Real] | None = None
            norm_msg: list[tuple[Any, Any]] | None = None
            err: Exception | None = None
            err_trace: str | None = None

            # Handle all exceptions
            try:
                scores = self._get_score(data_item, **kwargs)
                eval_scores = self._check_scores(scores)
                norm_msg = [
                    (f"[{prefix}/ IT]", format_data_item(data_item)),
                    (f"[{prefix}/ SC]", eval_scores),
                ]
            except Exception as e:
                err, err_trace = e, traceback.format_exc()

            # Write the log on demand
            if (
                self.logging_mode == "failed"
                and eval_scores is None
                or self.logging_mode == "all"
            ):
                mlog_item = {
                    "time": str(datetime.now()),
                    "index": i,
                    "iter": it,
                    "eval_scores": eval_scores,
                    "norm_msg": str(norm_msg),
                    "err_msg": err_trace,
                }
                with open(mlog_path, "a", encoding="utf-8") as f:
                    f.write(json.dumps(mlog_item, ensure_ascii=False) + "\n")

            # Return the information based on success or failure
            if eval_scores is not None and norm_msg is not None:
                return (i, it, eval_scores), norm_msg, None
            return None, None, (f"[{prefix}/ ERR]", f"{type(err).__name__}: {err}")

        async def process_afunc(
            item: tuple[int, int, DataItemType],
            addtlks: list[asyncio.Lock] | None = None,
            **kwargs,
        ) -> tuple[
            tuple[int, int, dict[Any, Real]], list[tuple[Any, Any]], None
        ] | tuple[None, None, tuple[Any, Any]]:
            """The asynchronous processing function."""
            i, it, data_item = item
            prefix = f"{i}" if self.n_iter == 1 else f"{i}.{it}"
            eval_scores: dict[Any, Real] | None = None
            norm_msg: list[tuple[Any, Any]] | None = None
            err: Exception | None = None
            err_trace: str | None = None

            # Handle all exceptions
            try:
                scores = await self._aget_score(data_item, **kwargs)
                eval_scores = self._check_scores(scores)
                norm_msg = [
                    (f"[{prefix}/ IT]", format_data_item(data_item)),
                    (f"[{prefix}/ SC]", eval_scores),
                ]
            except Exception as e:
                err, err_trace = e, traceback.format_exc()

            # Write the log on demand
            if (
                self.logging_mode == "failed"
                and eval_scores is None
                or self.logging_mode == "all"
            ):
                mlog_item = {
                    "time": str(datetime.now()),
                    "index": i,
                    "iter": it,
                    "eval_scores": eval_scores,
                    "norm_msg": str(norm_msg),
                    "err_msg": err_trace,
                }
                assert (
                    isinstance(addtlks, list) and len(addtlks) >= 1
                ), "no lock for log in async mode; this is most likely an ml3m bug."
                async with addtlks[0]:
                    with open(mlog_path, "a", encoding="utf-8") as f:
                        f.write(json.dumps(mlog_item, ensure_ascii=False) + "\n")

            # Return the information based on success or failure
            if eval_scores is not None and norm_msg is not None:
                return (i, it, eval_scores), norm_msg, None
            return None, None, (f"[{prefix}/ ERR]", f"{type(err).__name__}: {err}")

        # Activate the asynchronous runner (sequential mode if only one worker)
        runner = AsyncRunner(
            process_func=process_func,
            process_afunc=process_afunc,
            verbose=self.verbose,
        )
        results: list[tuple[int, int, dict[Any, Real]]]
        failed: list[tuple[int, int, DataItemType]]
        results, failed = runner.run(
            items=self._yield_dataset(overwrite=overwrite),
            worker_kwargs=self._worker_kwargs,
            n_locks=1,
        )
        completed = len(failed) == 0

        # Aggregate the result if needed
        new_df: pd.DataFrame
        failed_indices = [item[0] for item in failed]
        if self.n_iter == 1:
            i2scores = {i: scores for i, _, scores in results}
            new_df = pd.DataFrame(i2scores).T
        else:
            iit2scores = {(i, it): scores for i, it, scores in results}
            grp = pd.DataFrame(iit2scores).T.groupby(level=0)
            if self.agg_method in ["mean", "sum", "min", "max"]:
                new_df = grp.agg(self.agg_method)
            else:  # self.agg_method == "mode":
                # TODO: maybe resolve: error: "Iterable[Any]" has no attribute "mode"
                new_df = grp.apply(
                    lambda x: x.mode().iloc[0]  # type: ignore[attr-defined]
                )

        # Update the file with the obtained results
        for i, row in new_df.iterrows():
            # TODO: maybe resolve: error: Invalid index type "tuple[Hashable | None,
            # Index]" for "_LocIndexerFrame"; expected type ...
            # not sure whether the current operation is perfectly valid in all cases
            self._result_df.loc[i, self.subjects] = row  # type: ignore[index]
        for i in failed_indices:
            self._result_df.loc[i, self.subjects] = None
        self._result_df = self._result_df.convert_dtypes().sort_index()
        self._result_df.reset_index(names="i").to_csv(self.save_path, index=False)

        # Summarize the save location (and possibly log location)
        print(colored("Results can be found at:", COLOR.GREEN))
        print(os.path.abspath(self.save_path))
        if self.logging_mode != "none" and os.path.exists(mlog_path):
            print(colored("Execution log can be found at:", COLOR.GREEN))
            print(os.path.abspath(os.path.abspath(mlog_path)))
        return completed

    def load_scores(
        self,
        subject_subset: list | None = None,
        items: list | None = None,
    ) -> pd.DataFrame:
        """Load the scores from the save location.

        Parameters
        ----------
        subject_subset : list or None
            The subjects of the scores to select, i.e., the columns. If ``None``,
            select all subjects. In the returned :class:`pd.DataFrame`, the columns
            will be in the same order as ``subject_subset``.
        items : list or None
            The indices of the items to select. If ``None``, select all items. This
            will be applied after ``subject_subset``. This does not necessarily need
            to be a subset of the index of the loaded :class:`pd.DataFrame`. The
            indices that do not exist in the index of the loaded :class:`pd.DataFrame`
            will be assigned ``NaN``. In the returned :class:`pd.DataFrame`, the rows
            will be in the same order as ``items``.

        Returns
        -------
        scores : pandas.DataFrame
            The scores loaded from ``save_path``.

        Examples
        --------
        Suppose that the file at ``save_path`` looks like the following:

        .. code-block::

            i,score1,score2
            0,78,83
            1,64,76
            3,100,92
            4,28,38
            5,30,45

        >>> evaluator.load_scores()  # doctest: +SKIP
           score1  score2
        i
        0      78      83
        1      64      76
        3     100      92
        4      28      38
        5      30      45

        >>> evaluator.load_scores(subject_subset=["score2"])  # doctest: +SKIP
           score2
        i
        0      83
        1      76
        3      92
        4      38
        5      45

        >>> evaluator.load_scores(items=list(range(7)))  # doctest: +SKIP
           score1  score2
        i
        0    78.0    83.0
        1    64.0    76.0
        2     NaN     NaN
        3   100.0    92.0
        4    28.0    38.0
        5    30.0    45.0
        6     NaN     NaN
        """
        result_df = pd.read_csv(self.save_path).set_index("i")
        if subject_subset is not None:
            result_df = result_df[subject_subset]
        if items is not None:
            result_df = result_df.reindex(items)
        return result_df

    def load_avg_score(
        self,
        subject_subset: list | None = None,
        items: list | None = None,
    ) -> dict[str, Real]:
        """Load the average score of each subject from the save location.

        Parameters
        ----------
        subject_subset : list or None
            The subjects of the scores to select, i.e., the columns. If ``None``,
            select all subjects.
        items : list or None
            The indices of the items to select. If ``None``, select all items. This
            will be applied after ``subject_subset``. This does not necessarily need
            to be a subset of the index of the loaded :class:`pd.DataFrame`. However,
            any item out of range would not be taken into account when computing the
            average score.

        Returns
        -------
        avg_score : pandas.DataFrame
            The average score loaded from ``save_path``.

        Examples
        --------
        Suppose that the file at ``save_path`` looks like the following:

        .. code-block::

            i,score1,score2
            0,78,83
            1,64,76
            2,100,92
            3,28,38
            4,30,45

        >>> evaluator.load_avg_score()  # doctest: +SKIP
        {'score1': 60.0, 'score2': 66.8}

        >>> evaluator.load_avg_score(subject_subset=["score2"])  # doctest: +SKIP
        {'score2': 66.8}

        >>> evaluator.load_avg_score(items=list(range(7)))  # doctest: +SKIP
        {'score1': 60.0, 'score2': 66.8}
        """
        result_df = self.load_scores(subject_subset=subject_subset, items=items)
        return result_df.mean().to_dict()


class BaseOpenAIEvaluator(BaseEvaluator):
    """Base evaluator class via OpenAI.

    .. note::
        This class is meant to be subclassed. The methods that must be overriden
        include:

        - :meth:`BaseOpenAIEvaluator._prompt`
        - :meth:`BaseOpenAIEvaluator._extract_scores`

    Parameters
    ----------
    dataset : str or pathlib.Path
        The absolute path to the evaluation dataset.
    save_path : str or pathlib.Path
        The absolute path to the save location. This path may or may not exist, and if
        it exists, its file contents will be treated as a (partially) written result.
        Whether to overwrite the existing results or to build on them depend on
        ``overwrite`` when using the :meth:`BaseOpenAIEvaluator.evaluate` method.
    subjects : list
        The subjects to evaluate. This should *strictly* correspond to how to scores
        are obtained in :meth:`BaseOpenAIEvaluator._extract_scores`. If the score is
        obtained as a single real value, ``subjects`` must be a list of one element and
        that element will be used as the name of that score. If the score(s) are
        obtained as a dictionary of subject-value pairs, all items in ``subjects`` must
        appear as the keys of the obtained dictionary. Any additional key will be
        discarded; any missing key will be treated as an error.
    openai_config : str or pathlib.Path
        The absolute path to the OpenAI configuration file.
    fmt : {"jsonl", "json", "csv"}, default="jsonl"
        The format of ``dataset``.
    n_iter : int, default=1
        The number of iterations for each data item. This is commonly used when one
        round of scoring is not convincing enough and the final scoring should be some
        statistics of multiple rounds of scoring.
    agg_method : {"mean", "sum", "min", "max", "mode"}, default=None
        The aggregate method to use on multiple rounds of scoring. Ignored when
        ``n_iter=1``, and otherwise this must not be ``None``.
    timeout : float, default=60
        The timeout in seconds. This is not the OpenAI timeout, but the timeout for
        cancelling the worker tasks.
    model : str, default="gpt-3.5-turbo"
        The ID of the model to use, must be one of the available OpenAI models that
        support the ChatCompletion API. See also
        https://platform.openai.com/docs/models/model-endpoint-compatibility.
    logging_mode : {"all", "failed", "none"}, default="all"
        The logging mode, whether to save the logs of all items, or only of failed
        items, or save no log.
    verbose : int, default=0
        The verbosity level of the processing. For negative levels, only a progress bar
        will be displayed. For level 0, the errored items will also be displayed. For
        positive levels, the all items will be displayed, and the verbosity level
        determines the number of lines to display for the message of each item.
    openai_kwargs
        The additional keyword arguments to pass to OpenAI ChatCompletion. See the
        arguments marked as *Optional* at
        https://platform.openai.com/docs/api-reference/chat/create. Do not try to pass
        ``api_key`` and ``api_base`` through ``openai_kwargs``, use a configuration
        file instead.
    """

    def __init__(
        self,
        dataset: str | Path,
        save_path: str | Path,
        subjects: list,
        openai_config: str | Path,
        *,
        fmt: DatasetFormat = "jsonl",
        n_iter: int = 1,
        agg_method: AggregateMethod | None = None,
        timeout: float = 60,
        model: str = "gpt-3.5-turbo",
        logging_mode: LoggingMode = "all",
        verbose: int = 0,
        **openai_kwargs,
    ) -> None:
        self.openai_config = openai_config
        self.timeout = timeout
        self.model = model
        self.openai_kwargs = openai_kwargs

        # Load the OpenAI configurations
        validate_path(self.openai_config)
        worker_kwargs = [
            {"api_key": config.key, "api_base": config.base}
            for config in get_openai_config(self.openai_config)
            for _ in range(config.n_workers)
        ]
        if "api_key" in self.openai_kwargs or "api_base" in self.openai_kwargs:
            raise InvalidParameterError(
                "**openai_kwargs",
                actual=self.openai_kwargs,
                reason=(
                    "'api_key' and 'api_base' should be specified with a "
                    "configuration file"
                ),
            )

        # Inherit from parent
        super().__init__(
            dataset=dataset,
            save_path=save_path,
            subjects=subjects,
            fmt=fmt,
            workers=worker_kwargs,
            n_iter=n_iter,
            agg_method=agg_method,
            logging_mode=logging_mode,
            verbose=verbose,
        )

    def _prompt(self, data_item: DataItemType) -> tuple[str, str]:
        """Return the prompt for evaluation.

        :meta public:

        .. note::
            This method is not implemented and must be overridden in subclasses.

        Parameters
        ----------
        data_item : DataItemType
            The data item.

        Returns
        -------
        sys_msg : str
            The system message for setting the role of the OpenAI model when querying
            for evaluation, e.g. a professional teacher in some field. If no system
            message is needed, this should be an empty string. See also
            https://platform.openai.com/docs/guides/gpt/chat-completions-api
            for an example of system message.
        eval_prompt : str
            The formatted evaluation prompt.
        """
        raise NotImplementedError

    def _extract_scores(
        self, reply: str, data_item: DataItemType
    ) -> Real | dict[Any, Real]:
        """Extract the score(s) from the OpenAI model reply (and the data item).

        :meta public:

        .. note::
            This method is not implemented and must be overridden in subclasses.

        Parameters
        ----------
        reply : str
            The OpenAI model reply, from which the score(s) will be extracted.
        data_item : DataItemType
            The data item. This may not be used, but in case the OpenAI model reply
            requires comparion with the data item to give the final score, the data
            item is passed in as well.

        Returns
        -------
        scores : real or dict
            The extracted scores, either a single score or a dictionary of subject-
            score pairs.

        Notes
        -----
        This method should correspond to the :meth:`BaseOpenAIEvaluator._prompt`
        method, in the sense that the formatted evaluation prompt is expected to invoke
        an *extractable* model reply, and this method should extract the score(s) from
        that reply. It can extract either a single score or a dictionary of subject-
        score pairs.

        Note that if there are mutliple scores obtained, i.e., returning a dictionary,
        remember that the "i" key cannot be included since it is reserved for indexing.

        It is recommended *not* to catch the exceptions that cause the extraction of
        scores to fail, since otherwise :meth:`BaseOpenAIEvaluator.evaluate` will not
        realize that the data item errors out.
        """
        raise NotImplementedError

    def _prior_scoring(self, data_item: DataItemType):
        """Get messages prior to OpenAI querying.

        Parameters
        ----------
        data_item : DataItemType
            The data item.

        Returns
        -------
        messages : list of dict
            The messages for OpenAI querying.
        """
        sys_msg, eval_prompt = self._prompt(data_item)
        return (
            [
                {"role": "system", "content": sys_msg},
                {"role": "user", "content": eval_prompt},
            ]
            if sys_msg
            else [{"role": "user", "content": eval_prompt}]
        )

    def _post_scoring(
        self, completion: Any, data_item: DataItemType
    ) -> Real | dict[Any, Real]:
        """Process the OpenAI response posterior to querying.

        Parameters
        ----------
        completion : an OpenAI completion
            The returned OpenAI completion.

        Returns
        -------
        scores : real or dict
            The extracted scores, either a single score or a dictionary of subject-
            score pairs.
        """
        finish_reason = completion["choices"][0]["finish_reason"]
        if finish_reason != "stop":
            raise ValueError(f"Model terminated by '{finish_reason}'")
        return self._extract_scores(
            completion["choices"][0]["message"]["content"], data_item
        )

    def _get_score(self, data_item: DataItemType, **kwargs) -> Real | dict[Any, Real]:
        """:meta private:"""
        messages = self._prior_scoring(data_item)
        completion = openai.ChatCompletion.create(
            model=self.model,
            messages=messages,
            api_key=kwargs["api_key"],
            api_base=kwargs["api_base"],
            **self.openai_kwargs,
        )
        return self._post_scoring(completion, data_item)

    async def _aget_score(
        self, data_item: DataItemType, **kwargs
    ) -> Real | dict[Any, Real]:
        """:meta private:"""
        messages = self._prior_scoring(data_item)
        completion = await asyncio.wait_for(
            openai.ChatCompletion.acreate(
                model=self.model,
                messages=messages,
                api_key=kwargs["api_key"],
                api_base=kwargs["api_base"],
                **self.openai_kwargs,
            ),
            timeout=self.timeout,
        )
        return self._post_scoring(completion, data_item)



import asyncio
import json
import os
import random
import re
import threading
from itertools import product

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal, assert_series_equal

from ml3m.base import BaseEvaluator, BaseOpenAIEvaluator
from ml3m.errors import InvalidParameterError, ScoringError

random.seed(2023)


#######################################################################################
#                                                                                     #
#                                        DATA                                         #
#                                                                                     #
#######################################################################################


dataset_2 = [
    {
        "instruction": "What is the capital of China?",
        "input": "",
        "output": "The capital of China is Beijing.",
        "response": "Beijing.",  # This is clearly correct
    },
    {
        "instruction": "What is the capital of France?",
        "input": "",
        "output": "The capital of France is Paris.",
        "response": "Marseille.",  # And this is clearly incorrect
    },
]

result_df_5_2 = pd.DataFrame(
    data=np.array([[78, 83], [64, 76], [100, 92], [28, 38], [30, 45]]),
    index=pd.Index([0, 1, 3, 4, 5], name="i"),
    columns=["score1", "score2"],
)

openai_configuration = [{"key": "sk-xxx", "base": None, "n_workers": 30}]

openai_configuration_1w = [{"key": "sk-xxx", "base": None, "n_workers": 1}]


@pytest.fixture(scope="module")
def prepare(request, storage):
    """Make a temporary storage and clear it towards the end."""
    paths = {}

    # Make files for `dataset_2`
    for fmt in ["jsonl", "json", "csv"]:
        dataset = os.path.join(
            storage, f"dataset_2__{request.keywords.node.name}.{fmt}"
        )
        if fmt == "jsonl":
            with open(dataset, "w", encoding="utf-8") as f:
                for item in dataset_2:
                    f.write(json.dumps(item, ensure_ascii=False) + "\n")
        elif fmt == "json":
            with open(dataset, "w", encoding="utf-8") as f:
                json.dump(dataset_2, f, ensure_ascii=False, indent=4)
        else:  # fmt == "csv"
            df = pd.DataFrame(dataset_2)
            df.to_csv(dataset, index=False)
        paths[f"dataset_2__{fmt}"] = dataset

    # Make file for `result_df_5_2`
    save_path = os.path.join(
        storage, f"result_df_5_2__{request.keywords.node.name}.csv"
    )
    result_df_5_2.reset_index(names="i").to_csv(save_path, index=False)
    paths["result_df_5_2"] = save_path

    # Make file for `openai_configuration`
    save_path = os.path.join(
        storage, f"openai_configuration__{request.keywords.node.name}.json"
    )
    with open(save_path, "w", encoding="utf-8") as f:
        json.dump(openai_configuration, f, indent=4, ensure_ascii=False)
    paths["openai_configuration"] = save_path

    # Make file for `openai_configuration_1w`
    save_path = os.path.join(
        storage, f"openai_configuration_1w__{request.keywords.node.name}.json"
    )
    with open(save_path, "w", encoding="utf-8") as f:
        json.dump(openai_configuration_1w, f, indent=4, ensure_ascii=False)
    paths["openai_configuration_1w"] = save_path

    return paths


#######################################################################################
#                                                                                     #
#                                  PREPARATION WORK                                   #
#                                                                                     #
#######################################################################################


class NormalBaseEvaluator(BaseEvaluator):
    """Extends the base evaluator.

    This can return random scores, fail on all items, or fail on a certain item based
    on the instruction field.
    """

    def __init__(
        self,
        dataset,
        save_path,
        subjects,
        *,
        fmt="jsonl",
        workers=1,
        n_iter=1,
        agg_method=None,
        logging_mode="all",
        verbose=1,
        mode="random",
    ):
        super().__init__(
            dataset,
            save_path,
            subjects,
            fmt=fmt,
            workers=workers,
            n_iter=n_iter,
            agg_method=agg_method,
            logging_mode=logging_mode,
            verbose=verbose,
        )
        self.mode = mode

    def _get_score(self, data_item, **kwargs):
        if self.mode == "random" or self.mode.startswith("err_on_"):
            if (
                self.mode.startswith("err_on_instruction_")
                and data_item["instruction"] == self.mode[19:]
            ):
                raise ValueError
            if len(self.subjects) == 1:
                return random.randint(0, 100)
            return {subject: random.randint(0, 100) for subject in self.subjects}
        elif self.mode == "all_err":
            raise ValueError

    async def _aget_score(self, data_item, **kwargs):
        await asyncio.sleep(0.01)
        return self._get_score(data_item, **kwargs)


class MultIterBaseEvaluator(BaseEvaluator):
    """Extends the base evaluator.

    This supports taking a function ``sc_mapping`` that maps the iteration index and
    the subject index to a score. It does not tell different data items apart.
    """

    def __init__(
        self,
        dataset,
        save_path,
        subjects,
        sc_mapping,
        *,
        fmt="jsonl",
        workers=1,
        n_iter=1,
        agg_method=None,
        logging_mode="all",
        verbose=1,
    ):
        super().__init__(
            dataset,
            save_path,
            subjects,
            fmt=fmt,
            workers=workers,
            n_iter=n_iter,
            agg_method=agg_method,
            logging_mode=logging_mode,
            verbose=verbose,
        )
        self.sc_mapping = sc_mapping
        self.lock = threading.Lock()
        self.tracker = {}

    def _get_score(self, data_item, **kwargs):
        instruction = data_item["instruction"]
        with self.lock:
            if instruction in self.tracker:
                ind = self.tracker[instruction]
                self.tracker[instruction] += 1
            else:
                ind = 0
                self.tracker[instruction] = 1
        return {
            subject: self.sc_mapping(ind, i) for i, subject in enumerate(self.subjects)
        }

    async def _aget_score(self, data_item, **kwargs):
        await asyncio.sleep(0.01)
        return self._get_score(data_item, **kwargs)


class NormalBaseOpenAIEvaluator(BaseOpenAIEvaluator):
    """Extends the base OpenAI evaluator.

    This can judges the correctness of response based on whether the response is ``in``
    the reference answer.
    """

    _pattern = re.compile(r"```\n(.+)\n```\n\n```\n(.+)\n```", re.DOTALL)

    def __init__(
        self,
        dataset,
        save_path,
        subjects,
        openai_config,
        *,
        fmt="jsonl",
        n_iter=1,
        agg_method=None,
        timeout=60,
        model="gpt-3.5-turbo",
        logging_mode="all",
        verbose=1,
        **openai_kwargs,
    ):
        super().__init__(
            dataset,
            save_path,
            subjects,
            openai_config,
            fmt=fmt,
            n_iter=n_iter,
            agg_method=agg_method,
            timeout=timeout,
            model=model,
            logging_mode=logging_mode,
            verbose=verbose,
            **openai_kwargs,
        )

    def _prompt(self, data_item):
        reference, actual = data_item["output"], data_item["response"]
        return "", f"```\n{reference}\n```\n\n```\n{actual}\n```"

    def _extract_scores(self, reply, data_item):
        return int(reply)


def mock_openai_chatcompletion_create(*args, **kwargs):
    """Mock function for openai.ChatCompletion.create."""
    for item in kwargs["messages"]:
        if item["role"] == "user":
            target_msg = item["content"]
            break

    mat = re.match(r"```\n(.+)\n```\n\n```\n(.+)\n```", target_msg, re.DOTALL)
    reference, actual = mat.group(1), mat.group(2)
    score = (actual in reference) * 100

    return {
        "choices": [
            {
                "finish_reason": "stop",
                "message": {"content": str(score), "role": "assistant"},
            }
        ],
        "usage": {f"{term}_tokens": 1 for term in ["completion", "prompt", "total"]},
    }


async def mock_openai_chatcompletion_acreate(*args, **kwargs):
    """Mock function for openai.ChatCompletion.acreate."""
    await asyncio.sleep(0.01)
    return mock_openai_chatcompletion_create(*args, **kwargs)


def mock_openai_chatcompletion_create_fail2stop(*args, **kwargs):
    """Mock function for openai.ChatCompletion.create."""
    return {
        "choices": [
            {
                "finish_reason": "null",
                "message": {"content": "Hello.", "role": "assistant"},
            }
        ],
        "usage": {f"{term}_tokens": 1 for term in ["completion", "prompt", "total"]},
    }


async def mock_openai_chatcompletion_acreate_fail2stop(*args, **kwargs):
    """Mock function for openai.ChatCompletion.acreate."""
    await asyncio.sleep(0.01)
    return mock_openai_chatcompletion_create_fail2stop(*args, **kwargs)


#######################################################################################
#                                                                                     #
#                                  TESTS START HERE                                   #
#                                                                                     #
#######################################################################################


class TestBaseEvaluator:
    """Testing ml3m.base.BaseEvaluator."""

    @pytest.mark.parametrize("fmt", ["jsonl", "json", "csv"])
    @pytest.mark.parametrize("subjects", [["score1"], ["score2", "score3"]])
    @pytest.mark.parametrize("workers", [1, 3])
    @pytest.mark.parametrize("n_iter,agg_method", [(1, None), (3, "sum"), (3, "mode")])
    @pytest.mark.parametrize("logging_mode", ["none", "all", "failed"])
    @pytest.mark.parametrize("verbose", [0, 1, 2, 3])
    def test_base_evaluator_result_versus_written(
        self,
        fmt,
        subjects,
        workers,
        n_iter,
        agg_method,
        logging_mode,
        verbose,
        storage,
        prepare,
        request,
    ):
        """Test that evaluator._result_df and the written csv are the same.

        This serves as a basis so that the other tests do not need to I/O to check the
        results but instead directly read the attribute.
        """
        dataset = prepare[f"dataset_2__{fmt}"]
        save_path = os.path.join(
            storage, f"save_path__{request.keywords.node.name}.csv"
        )

        evaluator = NormalBaseEvaluator(
            dataset=dataset,
            save_path=save_path,
            subjects=subjects,
            fmt=fmt,
            workers=workers,
            n_iter=n_iter,
            agg_method=agg_method,
            logging_mode=logging_mode,
            verbose=verbose,
        )
        completed = evaluator.evaluate()
        assert completed

        df_saved = pd.read_csv(save_path)
        df_stored = evaluator._result_df.reset_index(names="i")
        assert_frame_equal(df_saved, df_stored, check_dtype=False)

    @pytest.mark.parametrize("fmt", ["jsonl", "json", "csv"])
    @pytest.mark.parametrize("subjects", [["score1"], ["score2", "score3"]])
    @pytest.mark.parametrize("new_subjects", [["score4"], ["score5", "score6"]])
    @pytest.mark.parametrize("workers", [1, 3])
    def test_base_evaluator_evaluate_basics(
        self, fmt, subjects, new_subjects, workers, storage, prepare, request
    ):
        """Test the basic evaluator functionalities.

        Fail all data items
        -> Pass one of the data items
        -> Pass all data items
        -> Evaluate again (should make no change)
        -> Fail all data items on the new subjects (should make no change)
        -> Pass one of the data items on the new subjects
        -> Evaluate on old subject(s) again (should make no change)
        -> Pass all data items on the new subjects
        -> Evaluate again (should make no change)
        """
        dataset = prepare[f"dataset_2__{fmt}"]
        save_path = os.path.join(
            storage, f"save_path__{request.keywords.node.name}.csv"
        )

        # This should pass none of the data items
        evaluator = NormalBaseEvaluator(
            dataset=dataset,
            save_path=save_path,
            subjects=subjects,
            fmt=fmt,
            workers=workers,
            mode="all_err",
        )
        completed = evaluator.evaluate()
        assert not completed

        df = evaluator._result_df
        assert list(df.index) == [0, 1]
        assert list(df.columns) == subjects
        assert df.isna().all().all()

        # This should pass the second data item but fail the first
        item_0_instruction = dataset_2[0]["instruction"]
        evaluator.mode = f"err_on_instruction_{item_0_instruction}"
        completed = evaluator.evaluate()
        assert not completed

        df = evaluator._result_df
        item_1_scores = df.loc[1, :]
        assert not item_1_scores.isna().any()

        # This should pass all of the data items
        evaluator.mode = "random"
        completed = evaluator.evaluate()
        assert completed

        df = evaluator._result_df
        items_scores = df.copy()
        assert not df.isna().any().any()
        assert_series_equal(item_1_scores, df.loc[1, :])

        # This should not modify the evaluation results
        evaluator.evaluate()
        completed = evaluator.evaluate()
        assert completed

        df = evaluator._result_df
        assert_frame_equal(items_scores, df)

        # This should pass none of the data items on the new subjects
        evaluator2 = NormalBaseEvaluator(
            dataset=dataset,
            save_path=save_path,
            subjects=new_subjects,
            fmt=fmt,
            workers=workers,
            mode="all_err",
        )
        completed = evaluator2.evaluate()
        assert not completed

        df = evaluator2._result_df
        assert list(df.columns) == [*subjects, *new_subjects]
        assert df.loc[:, new_subjects].isna().all().all()
        assert_frame_equal(items_scores, df.loc[:, subjects])

        # This should pass the second data item but fail the first on the new subjects
        evaluator2.mode = f"err_on_instruction_{item_0_instruction}"
        completed = evaluator2.evaluate()
        assert not completed

        df = evaluator2._result_df
        item_1_scores = df.loc[1, :]
        new_items_scores = df.copy()
        assert df.loc[0, new_subjects].isna().all()
        assert not df.loc[1, new_subjects].isna().any()
        assert_frame_equal(items_scores, df[subjects])

        # This should not modify the evaluation results since it is using old subjects
        completed = evaluator.evaluate()
        assert completed

        df = evaluator._result_df
        assert_frame_equal(new_items_scores, df)

        # This should pass all of the data items on the new subjects
        evaluator2.mode = "random"
        completed = evaluator2.evaluate()
        assert completed

        df = evaluator2._result_df
        new_items_scores = df.copy()
        assert list(df.index) == [0, 1]
        assert list(df.columns) == [*subjects, *new_subjects]
        assert not df.isna().any().any()
        assert_series_equal(item_1_scores, df.loc[1, :])

        # This should not modify the evaluation results
        evaluator2.evaluate()
        completed = evaluator2.evaluate()
        assert completed

        df = evaluator2._result_df
        assert_frame_equal(new_items_scores, df)

    @pytest.mark.parametrize("fmt", ["jsonl", "json", "csv"])
    @pytest.mark.parametrize("subjects", [["score1"], ["score2", "score3"]])
    @pytest.mark.parametrize("new_subjects", [["score4"], ["score5", "score6"]])
    @pytest.mark.parametrize("workers", [1, 3])
    def test_base_evaluator_overwrite(
        self, fmt, subjects, new_subjects, workers, storage, prepare, request
    ):
        """Test overwrite parameter of evaluate.

        Pass all data items
        -> Overwrite and fail all data items
        -> Overwrite and pass all data items
        -> Overwrite and pass all data items on the new subjects
        -> Overwrite and fail all data items on the new subjects
        -> Pass all data items on the new subjects
        -> Overwrite and pass all data items on the new subjects
        """
        dataset = prepare[f"dataset_2__{fmt}"]
        save_path = os.path.join(
            storage, f"save_path__{request.keywords.node.name}.csv"
        )

        # Preparation: Pass all data items
        evaluator = NormalBaseEvaluator(
            dataset=dataset,
            save_path=save_path,
            subjects=subjects,
            fmt=fmt,
            workers=workers,
        )
        evaluator.evaluate()
        df_ref = evaluator._result_df

        # Overwrite and fail all data items
        evaluator.mode = "all_err"
        completed = evaluator.evaluate(overwrite=True)
        assert not completed

        df = evaluator._result_df
        assert df.isna().all().all()

        # Overwrite and pass all data items
        evaluator.mode = "random"
        evaluator.evaluate(overwrite=True)
        df_ref = evaluator._result_df

        # Overwrite and pass all data items on the new subjects
        completed = evaluator.evaluate(overwrite=True)
        assert completed

        df = evaluator._result_df
        with pytest.raises(AssertionError):
            assert_frame_equal(df, df_ref)
        df_ref = df

        # Overwrite and fail all data items on the new subjects
        evaluator2 = NormalBaseEvaluator(
            dataset=dataset,
            save_path=save_path,
            subjects=new_subjects,
            fmt=fmt,
            workers=workers,
            mode="all_err",
        )
        completed = evaluator2.evaluate(overwrite=True)
        assert not completed

        df = evaluator2._result_df
        assert_frame_equal(df.loc[:, subjects], df_ref)
        assert df.loc[:, new_subjects].isna().all().all()

        # Preparation: Pass all data items on the new subjects
        evaluator2.mode = "random"
        evaluator2.evaluate()
        df_ref = evaluator2._result_df

        # Overwrite and pass all data items on the new subjects
        completed = evaluator2.evaluate(overwrite=True)
        assert completed

        df = evaluator2._result_df
        with pytest.raises(AssertionError):
            assert_frame_equal(df, df_ref)

    @pytest.mark.parametrize("fmt", ["jsonl", "json", "csv"])
    @pytest.mark.parametrize("workers", [1, 3])
    @pytest.mark.parametrize("agg_method", ["mean", "sum", "min", "max", "mode"])
    def test_base_evaluator_aggregate(
        self, fmt, workers, agg_method, storage, prepare, request
    ):
        """Test the aggregation methods."""
        subjects = ["score1", "score2"]
        dataset = prepare[f"dataset_2__{fmt}"]
        save_path = os.path.join(
            storage, f"save_path__{request.keywords.node.name}.csv"
        )

        evaluator = MultIterBaseEvaluator(
            dataset=dataset,
            save_path=save_path,
            subjects=subjects,
            sc_mapping=lambda i_iter, i_subject: i_iter % 4 + i_subject * 10,
            fmt=fmt,
            workers=workers,
            n_iter=5,
            agg_method=agg_method,
        )
        evaluator.evaluate()

        # For each data item, the scores would be
        # score1: 0, 1, 2, 3, 0
        # score2: 10, 11, 12, 13, 10

        if agg_method == "mean":
            # score1: (0 + 1 + 2 + 3 + 0) / 5 = 1.2
            # score2: (10 + 11 + 12 + 13 + 10) / 5 = 11.2
            results = [1.2, 11.2]
        elif agg_method == "sum":
            # score1: 0 + 1 + 2 + 3 + 0 = 6
            # score2: 10 + 11 + 12 + 13 + 10 = 56
            results = [6, 56]
        elif agg_method == "min":
            # score1: min(0, 1, 2, 3, 0) = 0
            # score2: min(10, 11, 12, 13, 10) = 10
            results = [0, 10]
        elif agg_method == "max":
            # score1: max(0, 1, 2, 3, 0) = 3
            # score2: max(10, 11, 12, 13, 10) = 13
            results = [3, 13]
        elif agg_method == "mode":
            # score1: mode(0, 1, 2, 3, 0) = 0
            # score2: mode(10, 11, 12, 13, 10) = 10
            results = [0, 10]

        expected = pd.DataFrame([results] * 2, columns=subjects)
        assert_frame_equal(
            evaluator._result_df, expected, check_dtype=False, check_names=False
        )

    @pytest.mark.parametrize("fmt", ["jsonl", "json", "csv"])
    def test_base_evalutor_overlapping_subjects(self, fmt, storage, prepare, request):
        """Test the behavior when two evaluations have overlapping subject.

        The overlapped subject will be overwritten without any warning or error.
        """
        dataset = prepare[f"dataset_2__{fmt}"]
        save_path = os.path.join(
            storage, f"save_path__{request.keywords.node.name}.csv"
        )

        evaluator = NormalBaseEvaluator(
            dataset=dataset,
            save_path=save_path,
            subjects=["score", "score1"],
            fmt=fmt,
        )
        completed = evaluator.evaluate()
        df = evaluator._result_df
        assert completed

        evaluator = NormalBaseEvaluator(
            dataset=dataset,
            save_path=save_path,
            subjects=["score"],
            fmt=fmt,
        )
        completed = evaluator.evaluate()
        assert completed
        assert_frame_equal(evaluator._result_df, df)

        evaluator2 = NormalBaseEvaluator(
            dataset=dataset,
            save_path=save_path,
            subjects=["score", "score2"],
            fmt=fmt,
        )
        completed = evaluator2.evaluate()
        df2 = evaluator2._result_df
        assert completed

        assert list(df2.columns) == ["score", "score1", "score2"]
        assert not df2.isna().any().any()

        # score1 will stay the same, but score should be overwritten even if overwrite
        # is not specified; this is because score2 is missing while score and score2
        # are treated as a whole in the second evaluator
        # Since the evaluator is in random mode, most probably the overwritten part
        # will not be the same as what it previously was
        assert_series_equal(df.loc[:, "score1"], df2.loc[:, "score1"])
        with pytest.raises(AssertionError):
            assert_series_equal(df.loc[:, "score"], df2.loc[:, "score"])

    def test_base_evalutor_invalid_init(self, storage, prepare, request):
        """Test invalid initialization."""
        dataset = prepare["dataset_2__jsonl"]
        save_path = os.path.join(
            storage, f"save_path__{request.keywords.node.name}.csv"
        )

        # Test invalid subjects
        for subjects in [[], "abc", ["i", "j"]]:
            with pytest.raises(InvalidParameterError):
                BaseEvaluator(
                    dataset=dataset,
                    save_path=save_path,
                    subjects=subjects,
                )

        # Test invalid fmt
        for fmt in [".csv", "txt"]:
            with pytest.raises(InvalidParameterError):
                BaseEvaluator(
                    dataset=dataset,
                    save_path=save_path,
                    subjects=["score"],
                    fmt=fmt,
                )

        # Test invalid logging_mode
        for logging_mode in ["any", "succeeded"]:
            with pytest.raises(InvalidParameterError):
                BaseEvaluator(
                    dataset=dataset,
                    save_path=save_path,
                    subjects=["score"],
                    logging_mode=logging_mode,
                )

        # Test invalid n_iter and agg_method
        for n_iter, agg_method in zip(
            [0, 2.4, "3", 2, 2, 2],
            ["mean", "mean", "mean", None, "cumsum"],
        ):
            with pytest.raises(InvalidParameterError):
                BaseEvaluator(
                    dataset=dataset,
                    save_path=save_path,
                    subjects=["score"],
                    n_iter=n_iter,
                    agg_method=agg_method,
                )

        # Test invalid workers
        for workers in [{"key": 0}, 0, 2.4, [0, 1]]:
            with pytest.raises(InvalidParameterError):
                BaseEvaluator(
                    dataset=dataset,
                    save_path=save_path,
                    subjects=["score"],
                    workers=workers,
                )

    def test_base_evalutor_check_score(self, storage, prepare, request):
        """Test the `_check_score` method."""
        dataset = prepare["dataset_2__jsonl"]
        save_path = os.path.join(
            storage, f"save_path__{request.keywords.node.name}.csv"
        )

        evaluator_1 = BaseEvaluator(
            dataset=dataset,
            save_path=save_path,
            subjects=["score"],
        )
        evaluator_3 = BaseEvaluator(
            dataset=dataset,
            save_path=save_path,
            subjects=["score1", "score2", "score3"],
        )

        # Test score not real or dict
        for evaluator, scores in product(
            [evaluator_1, evaluator_3], [None, complex(1, 2), pd.NA]
        ):
            msg = re.escape(
                "The scores must be either a real number or a dictionary with real "
                f"values; got '{scores}' of type '{type(scores)}' instead."
            )
            with pytest.raises(ScoringError, match=msg):
                evaluator._check_scores(scores)

        # Test score dict with not all values being real
        for invalid_val in [None, complex(1, 2), pd.NA]:
            msg = re.escape(
                "The scores must be either a real number or a dictionary with real "
                f"values; got a dictionary but there exists 'score: {invalid_val}' of "
                f"type '{type(invalid_val)}'."
            )
            with pytest.raises(ScoringError, match=msg):
                evaluator_1._check_scores({"score": invalid_val})
            msg = re.escape(
                "The scores must be either a real number or a dictionary with real "
                f"values; got a dictionary but there exists 'score2: {invalid_val}' of "
                f"type '{type(invalid_val)}'."
            )
            with pytest.raises(ScoringError, match=msg):
                evaluator_3._check_scores(
                    {"score1": 1, "score2": invalid_val, "score3": invalid_val}
                )

        # Test exception when single score but multiple subjects
        msg = re.escape("Got a single score but multiple subjects are assigned.")
        with pytest.raises(ScoringError, match=msg):
            evaluator_3._check_scores(10)

        # Test exception when missing required subjects
        msg = re.compile(r"Missing keys: '.+', '.+'\.")
        with pytest.raises(ScoringError, match=msg):
            evaluator_3._check_scores({"score2": 10, "score4": 10})

        # Test normal functionalities
        for scores in [10, {"score": 10}, {"score": 10, "extra_score": 10}]:
            assert evaluator_1._check_scores(scores) == {"score": 10}
        expected = {"score1": 10, "score2": 10, "score3": 10}
        scores = expected.copy()
        assert evaluator_3._check_scores(scores) == expected
        scores["score4"] = 10
        assert evaluator_3._check_scores(scores) == expected

    @pytest.mark.parametrize("subject_subset", [None, ["score2"], ["score2", "score1"]])
    @pytest.mark.parametrize("items", [None, [3, 5, 0], list(range(7))])
    def test_base_evalutor_scoring_methods(self, subject_subset, items, prepare):
        """Test the `load_scores` and `load_avg_score` methods."""
        dataset = prepare["dataset_2__jsonl"]
        save_path = prepare["result_df_5_2"]

        evaluator = BaseEvaluator(
            dataset=dataset,
            save_path=save_path,
            subjects=["score1", "score2"],
        )

        expected_df = result_df_5_2
        if subject_subset is not None:
            expected_df = expected_df[subject_subset]
        if items is not None:
            expected_df = expected_df.reindex(items)

        subjects = ["score1", "score2"] if subject_subset is None else subject_subset
        expected_scores = {
            subject: np.mean(expected_df.loc[:, subject].dropna())
            for subject in subjects
        }

        result_df = evaluator.load_scores(subject_subset=subject_subset, items=items)
        result_scores = evaluator.load_avg_score(
            subject_subset=subject_subset, items=items
        )

        assert_frame_equal(result_df, expected_df, check_dtype=False)
        assert expected_scores == result_scores


class TestBaseOpenAIEvaluator:
    """Testing ml3m.base.BaseOpenAIEvaluator."""

    @pytest.mark.parametrize(
        "config_name", ["openai_configuration", "openai_configuration_1w"]
    )
    def test_base_openai_evaluator_evaluate_basics(
        self, config_name, storage, prepare, mocker, request
    ):
        """Test the basic evaluator functionalities."""
        dataset = prepare["dataset_2__jsonl"]
        openai_config = prepare[config_name]
        save_path = os.path.join(
            storage, f"save_path__{request.keywords.node.name}.csv"
        )

        patcher_create = mocker.patch("ml3m.base.eval.openai.ChatCompletion.create")
        patcher_acreate = mocker.patch("ml3m.base.eval.openai.ChatCompletion.acreate")
        evaluator = NormalBaseOpenAIEvaluator(
            dataset=dataset,
            save_path=save_path,
            openai_config=openai_config,
            subjects=["score"],
        )

        # Patch to always fail with stop_reason != "stop"
        patcher_create.side_effect = mock_openai_chatcompletion_create_fail2stop
        patcher_acreate.side_effect = mock_openai_chatcompletion_acreate_fail2stop
        completed = evaluator.evaluate()
        assert not completed

        df = evaluator._result_df
        expected = pd.DataFrame(index=pd.Index([0, 1], name="i"), columns=["score"])
        expected["score"] = None
        assert_frame_equal(df, expected, check_dtype=False)

        # Patch to respond normally
        patcher_create.side_effect = mock_openai_chatcompletion_create
        patcher_acreate.side_effect = mock_openai_chatcompletion_acreate
        completed = evaluator.evaluate()
        assert completed

        df = evaluator._result_df
        expected.at[0, "score"] = 100
        expected.at[1, "score"] = 0
        assert_frame_equal(df, expected, check_dtype=False)

    def test_base_openai_evaluator_invalid_init(self, storage, prepare, request):
        """Test invalid initialization."""
        dataset = prepare["dataset_2__jsonl"]
        openai_config = prepare["openai_configuration"]
        save_path = os.path.join(
            storage, f"save_path__{request.keywords.node.name}.csv"
        )

        for openai_kwargs in [
            {"api_key": "xxx"},
            {"api_base": "xxx"},
            {"api_key": "xxx", "api_base": "xxx"},
        ]:
            with pytest.raises(InvalidParameterError):
                NormalBaseOpenAIEvaluator(
                    dataset=dataset,
                    save_path=save_path,
                    openai_config=openai_config,
                    subjects=["score"],
                    **openai_kwargs,
                )



"""This file is for logging-relevant functionalities."""


from __future__ import annotations

import os
from datetime import datetime

from ._paths import ensure_path


def manage_timed_logs(prefix, keep: int = 3) -> str:
    """Clean old logs and return the path to a new log.

    Parameters
    ----------
    prefix : str
        The prefix to the log.
    keep : int, default=3
        The number of logs to keep.

    Returns
    -------
    mlog_path : str
        The absolute path to the new OpenAI log. Note that the log is not created yet.
    """
    base = os.path.join(os.path.dirname(__file__), "logs")
    ensure_path(base, is_directory=True)
    dtformat = "%Y-%m-%d-%H.%M.%S.%f"
    fnames = [
        fname
        for fname in os.listdir(base)
        if fname.startswith(f"{prefix}_")
        and fname.endswith(".log")
        and os.path.isfile(os.path.join(base, fname))
    ]
    subs, sube = len(prefix) + 1, -4  # "{prefix}_" and ".log"
    fnames.sort(key=lambda fname: datetime.strptime(fname[subs:sube], dtformat))
    for fname in fnames[:-keep]:
        os.remove(os.path.join(base, fname))
    return os.path.join(base, f"{prefix}_{datetime.now().strftime(dtformat)}.log")



"""This file is for customizing exceptions."""


from __future__ import annotations

from typing import Any


class InvalidParameterError(Exception):
    """Exception for invalid parameters.

    Parameters
    ----------
    param_name: str
        The name of the invalid parameter.
    actual : Any, optional
        The actual value of the invalid parameter. If ``None``, this is not revealed in
        the error message.
    reason : str, optional
        The reason why the parameter is invalid. If ``None``, this is not revealed in
        the error message.
    """

    def __init__(
        self, param_name: str, *, actual: Any | None = None, reason: str | None = None
    ) -> None:
        msg = f"Invalid '{param_name}'"
        if actual is not None:
            msg += f"; got '{actual}'"
        if reason is not None:
            msg += f"; {reason}"
        msg += "."
        super().__init__(msg)


class ScoringError(Exception):
    """Exception when scoring.

    ``ScoringError`` can happen mainly in the following cases:

    - In an evaluator, the score obtained is not a real value, or the scores obtained
      are not represented as a dictionary with real values.
    - In an evaluator, there are multiple subjects assigned, but only a single score
      obtained.
    - In an evaluator, there exist assigned subjects that are not found in the obtained
      scores.
    """
